import os
from enum import Enum
from typing import List, Dict, Any, Optional, Union, Literal
from pathlib import Path
import pyterrier as pt


class TerrierModel(Enum):
    """A built-in Terrier weighting (scoring) model.

    This enum is primarily used with :meth:`TerrierIndex.retriever` to specify the weighting model to use.
    """
    bm25 = 'bm25'
    dph = 'dph'
    pl2 = 'pl2'
    dirichlet_lm = 'dirichlet_lm' # dirichletlm.mu=2500
    hiemstra_lm = 'hiemstra_lm' # hiemstra_lm.lambda=0.15
    tf = 'tf'
    tf_idf = 'tf_idf'


class TerrierIndex(pt.Artifact, pt.Indexer):
    """Represents a Terrier index.

    A Terrier index is a sparse inverted index structure that supports a variety of operations. It can be used to
    create transformers that perform retrieval, re-ranking, pseudo-relevance feedback, and other operations.
    """

    ARTIFACT_TYPE = 'sparse_index'
    ARTIFACT_FORMAT = 'terrier'
    ARTIFACT_PACKAGE_HINT = 'python-terrier'

    def __init__(
        self,
        path: Union[str, Path],
        *,
        memory: bool = False,
        _index_ref: object = None, 
        _index_obj: object = None
    ):
        """
        Args:
            path: The path to the index on disk.
            memory: Whether to load the index fully into memory.
            _index_ref: **For internal use only.** The Java IndexRef object for this index.
            _index_obj: **For internal use only.** The Java Index object for this index.
        """
        super().__init__(path)
        if _index_ref is not None:
            assert path is pt.Artifact.NO_PATH and _index_obj is None
        self._index_ref = _index_ref
        if _index_obj is not None:
            assert path is pt.Artifact.NO_PATH and _index_ref is None
        self._index_obj = _index_obj
        self.memory = memory


    # ----------------------------------------------------
    # Retrieval
    # ----------------------------------------------------

    def retriever(
        self,
        model: Union[TerrierModel, str],
        model_args: Optional[Dict[str, Any]] = None,
        *,
        num_results: int = 1000,
        include_fields: Optional[List[str]] = None,
        threads: int = 1,
        verbose: bool = False,
    ) -> pt.Transformer:
        """Creates a retriever transformer for this index.

        Args:
            model: The weighting model to use for scoring.
            model_args: The arguments to pass to the weighting model.
            num_results: The maximum number of results to return per query.
            include_fields: The metadata fields to return for each search result.
            threads: The number of threads to use during retrieval.
            verbose: Whether to progress information during retrieval

        Example Pipeline:

        .. schematic::
            :show_code:

            index = pt.terrier.TerrierIndex.example()
            # FOLD
            index.retriever('BM25', num_results=10)

        Terrier retrievers can also perform re-ranking when they receive a result frame as input:

        .. schematic::
            :show_code:
            :input_columns: qid,query,docno

            index = pt.terrier.TerrierIndex.example()
            # FOLD
            # As a re-ranker
            index.retriever('BM25', num_results=10)

        .. seealso::
            There are shorthand methods for creating common retrievers: :meth:`bm25`, :meth:`dph`.
        """
        if include_fields is None:
            include_fields = ['docno']
        else:
            if 'docno' not in include_fields:
                include_fields = ['docno'] + include_fields
        return pt.terrier.Retriever(
            self.index_obj(),
            controls=_map_controls(model_args),
            properties=_map_properties(model_args),
            metadata=include_fields,
            num_results=num_results,
            wmodel=_map_wmodel(model),
            threads=threads,
            verbose=verbose,
        )

    def bm25(
        self,
        *,
        k1: float = 1.2,
        b: float = 0.75,
        num_results: int = 1000,
        include_fields: Optional[List[str]] = None,
        threads: int = 1,
        verbose: bool = False,
    ) -> pt.Transformer:
        """Creates a BM25 retriever for this index.

        Args:
            k1: BM25's ``k1`` parameter, which controls TF saturation.
            b: BM25's ``b`` parameter, which controls the length penalty.
            num_results: The maximum number of results to return per query.
            include_fields: The metadata fields to return for each search result.
            threads: The number of threads to use during retrieval.
            verbose: Whether to progress information during retrieval

        Example Pipeline:

        .. schematic::
            :show_code:

            index = pt.terrier.TerrierIndex.example()
            # FOLD
            index.bm25()

        .. cite.dblp:: conf/trec/RobertsonWB98
        """
        return self.retriever(
            TerrierModel.bm25,
            {'bm25.k1': k1, 'bm25.b': b},
            num_results=num_results,
            include_fields=include_fields,
            threads=threads,
            verbose=verbose,
        )

    def dph(
        self,
        *,
        num_results: int = 1000,
        include_fields: Optional[List[str]] = None,
        threads: int = 1,
        verbose: bool = False,
    ) -> pt.Transformer:
        """Creates a DPH retriever for this index.

        Args:
            num_results: The maximum number of results to return per query.
            include_fields: The metadata fields to return for each search result.
            threads: The number of threads to use during retrieval.
            verbose: Whether to progress information during retrieval

        Example Pipeline:

        .. schematic::
            :show_code:

            index = pt.terrier.TerrierIndex.example()
            # FOLD
            index.dph()

        .. cite.dblp:: conf/ecir/Amati06
        """
        return self.retriever(
            TerrierModel.dph,
            num_results=num_results,
            include_fields=include_fields,
            threads=threads,
            verbose=verbose,
        )

    def pl2(
        self,
        *,
        c: float = 1.0,
        num_results: int = 1000,
        include_fields: Optional[List[str]] = None,
        threads: int = 1,
        verbose: bool = False,
    ) -> pt.Transformer:
        """Creates a PL2 retriever for this index.

        Args:
            c: PL2's ``c`` parameter, which controls the length normalization.
            num_results: The maximum number of results to return per query.
            include_fields: The metadata fields to return for each search result.
            threads: The number of threads to use during retrieval.
            verbose: Whether to progress information during retrieval

        Example Pipeline:

        .. schematic::
            :show_code:

            index = pt.terrier.TerrierIndex.example()
            # FOLD
            index.pl2()

        .. cite.dblp:: journals/tois/AmatiR02
        """
        return self.retriever(
            TerrierModel.pl2,
            {'dfr.c': c},
            num_results=num_results,
            include_fields=include_fields,
            threads=threads,
            verbose=verbose,
        )

    def dirichlet_lm(
        self,
        *,
        mu: float = 2500.0,
        num_results: int = 1000,
        include_fields: Optional[List[str]] = None,
        threads: int = 1,
        verbose: bool = False,
    ) -> pt.Transformer:
        """Creates a Dirichlet Language Model retriever for this index.

        Args:
            mu: Dirichlet LM's ``mu`` parameter, which controls the strength of the prior.
            num_results: The maximum number of results to return per query.
            include_fields: The metadata fields to return for each search result.
            threads: The number of threads to use during retrieval.
            verbose: Whether to progress information during retrieval

        Example Pipeline:

        .. schematic::
            :show_code:

            index = pt.terrier.TerrierIndex.example()
            # FOLD
            index.dirichlet_lm()

        .. cite.dblp:: journals/tois/ZhaiL04
        """
        return self.retriever(
            TerrierModel.dirichlet_lm,
            {'dirichletlm.mu': mu},
            num_results=num_results,
            include_fields=include_fields,
            threads=threads,
            verbose=verbose,
        )

    def hiemstra_lm(
        self,
        *,
        Lambda: float = 0.15,
        num_results: int = 1000,
        include_fields: Optional[List[str]] = None,
        threads: int = 1,
        verbose: bool = False,
    ) -> pt.Transformer:
        """Creates a Hiemstra Language Model retriever for this index.

        Args:
            Lambda: Hiemstra LM's ``lambda`` parameter, which controls the interpolation weight.
            num_results: The maximum number of results to return per query.
            include_fields: The metadata fields to return for each search result.
            threads: The number of threads to use during retrieval.
            verbose: Whether to progress information during retrieval

        Example Pipeline:

        .. schematic::
            :show_code:

            index = pt.terrier.TerrierIndex.example()
            # FOLD
            index.hiemstra_lm()

        .. cite.dblp:: phd/basesearch/Hiemstra01
        """
        return self.retriever(
            TerrierModel.hiemstra_lm,
            {'hiemstra_lm.lambda': Lambda},
            num_results=num_results,
            include_fields=include_fields,
            threads=threads,
            verbose=verbose,
        )

    def tf(
        self,
        *,
        num_results: int = 1000,
        include_fields: Optional[List[str]] = None,
        threads: int = 1,
        verbose: bool = False,
    ) -> pt.Transformer:
        """Creates a raw Term Frequency (TF) retriever for this index.

        This is typically useful for retrieving from learned sparse models.

        Args:
            num_results: The maximum number of results to return per query.
            include_fields: The metadata fields to return for each search result.
            threads: The number of threads to use during retrieval.
            verbose: Whether to progress information during retrieval

        Example Pipeline:

        .. schematic::
            :show_code:

            index = pt.terrier.TerrierIndex.example()
            # FOLD
            index.tf()
        """
        return self.retriever(
            TerrierModel.tf,
            num_results=num_results,
            include_fields=include_fields,
            threads=threads,
            verbose=verbose,
        )

    def tf_idf(
        self,
        *,
        k1: float = 1.2,
        b: float = 0.75,
        num_results: int = 1000,
        include_fields: Optional[List[str]] = None,
        threads: int = 1,
        verbose: bool = False,
    ) -> pt.Transformer:
        """Creates a TF-IDF retriever for this index.

        This retriever uses the Robertson formulation of TF and the Sparck Jones formulation of IDF.

        Args:
            k1: TF-IDF's ``k1`` parameter, which controls TF saturation.
            b: TF-IDF's ``b`` parameter, which controls the length penalty.
            num_results: The maximum number of results to return per query.
            include_fields: The metadata fields to return for each search result.
            threads: The number of threads to use during retrieval.
            verbose: Whether to progress information during retrieval

        Example Pipeline:

        .. schematic::
            :show_code:

            index = pt.terrier.TerrierIndex.example()
            # FOLD
            index.tf_idf()

        .. cite.dblp:: conf/trec/RobertsonWB98
        .. cite.dblp:: journals/jd/Jones04
        """
        return self.retriever(
            TerrierModel.tf_idf,
            {'tf_idf.k_1': k1, 'tf_idf.b': b},
            num_results=num_results,
            include_fields=include_fields,
            threads=threads,
            verbose=verbose,
        )


    # ----------------------------------------------------
    # Query Expansion / Rewriting
    # ----------------------------------------------------

    def rm3(
        self,
        *,
        fb_terms: int = 10,
        fb_docs: int = 3,
        fb_lambda: float = 0.6,
    ) -> pt.Transformer:
        """Creates an RM3 pseudo-relevance feedback transformer for this index.

        Args:
            fb_terms: The number of feedback terms to use.
            fb_docs: The number of feedback documents to use.
            fb_lambda: The interpolation weight between the original query and the feedback model.

        Example Pipeline:

        .. schematic::
            :show_code:

            index = pt.terrier.TerrierIndex.example()
            # FOLD
            index.bm25() >> index.rm3() >> index.bm25() >> pt.rewrite.reset()

        .. note::
            ``pt.rewrite.reset()`` is needed after the feedback step to reset the query to its original form.

        .. cite.dblp:: conf/trec/JaleelACDLLSW04
        """
        assert self.built(), f"{self!r} not built"
        assert self.index_obj().hasIndexStructure("direct"), f"{self!r} missing direct index"
        return pt.terrier.rewrite.RM3(
            self.index_obj(),
            fb_terms=fb_terms,
            fb_docs=fb_docs,
            fb_lambda=fb_lambda,
        )

    def bo1(
        self,
        *,
        fb_terms: int = 10,
        fb_docs: int = 3,
    ) -> pt.Transformer:
        """Creates a Bo1 pseudo-relevance feedback transformer for this index.

        Args:
            fb_terms: The number of feedback terms to use.
            fb_docs: The number of feedback documents to use.

        Example Pipeline:

        .. schematic::
            :show_code:

            index = pt.terrier.TerrierIndex.example()
            # FOLD
            index.bm25() >> index.bo1() >> index.bm25() >> pt.rewrite.reset()

        .. note::
            ``pt.rewrite.reset()`` is needed after the feedback step to reset the query to its original form.

        .. cite.dblp:: phd/ethos/Amati03
        """
        assert self.built(), f"{self!r} not built"
        assert self.index_obj().hasIndexStructure("direct"), f"{self!r} missing direct index"
        return pt.terrier.rewrite.Bo1QueryExpansion(
            self.index_obj(),
            fb_terms=fb_terms,
            fb_docs=fb_docs,
        )

    def kl(
        self,
        *,
        fb_terms: int = 10,
        fb_docs: int = 3,
    ) -> pt.Transformer:
        """Creates a KL-Divergence pseudo-relevance feedback transformer for this index.

        Args:
            fb_terms: The number of feedback terms to use.
            fb_docs: The number of feedback documents to use.

        Example Pipeline:

        .. schematic::
            :show_code:

            index = pt.terrier.TerrierIndex.example()
            # FOLD
            index.bm25() >> index.kl() >> index.bm25() >> pt.rewrite.reset()

        .. note::
            ``pt.rewrite.reset()`` is needed after the feedback step to reset the query to its original form.

        .. cite.dblp:: phd/ethos/Amati03
        """
        assert self.built(), f"{self!r} not built"
        assert self.index_obj().hasIndexStructure("direct"), f"{self!r} missing direct index"
        return pt.terrier.rewrite.KLQueryExpansion(
            self.index_obj(),
            fb_terms=fb_terms,
            fb_docs=fb_docs,
        )

    def sdm(self):
        """Creates a Sequential Dependence Model (SDM) query expansion transformer.

        Requires an that the index was built with positional information.

        .. cite.dblp:: conf/sigir/MetzlerC05
        """
        assert self.built(), f"{self!r} not built"
        assert self.collection_statistics().hasPositions(), f"{self!r} missing positional information"
        return pt.terrier.rewrite.SDM()


    # ----------------------------------------------------
    # Loading
    # ----------------------------------------------------

    def text_loader(
        self,
        fields: Union[List[str], str, Literal['*']] = '*',
        *,
        verbose=False,
    ) -> pt.Transformer:
        """Creates a transformer that loads stored text content from this index.

        Args:
            fields: The metadata fields to load for each document. If ``"*"``, loads all available fields.

        Example Pipeline:

        .. schematic::
            :show_code:

            index = pt.terrier.TerrierIndex.example()
            # FOLD
            index.text_loader()

        """
        return pt.terrier.TerrierTextLoader(self, fields, verbose=verbose)


    # ----------------------------------------------------
    # Indexing
    # ----------------------------------------------------

    def indexer(
        self,
        *,
        meta: Dict = {'docno': 20},
        text_attrs: List[str] = ['text'],
        tokeniser: Union[str, 'pt.terrier.TerrierTokeniser'] = 'english',
        stemmer: Union[None, str, 'pt.terrier.TerrierStemmer'] = 'porter',
        stopwords: Union[None, 'pt.terrier.TerrierStopwords', str, List[str]] = 'terrier',
        store_separate_fields: bool = False,
        store_positions: bool = False,
        threads: int = 1,
    ) -> pt.Indexer:
        """Returns an indexer that is used to build this index.

        Args:
            meta: The fields to store as metadata for each document. The keys are the metadata field names, and the values are the maximum lengths for each field.
            text_attrs: The text fields to index as text for each document.
            tokeniser: The tokeniser to use.
            stemmer: The stemmer to apply to each token.
            stopwords: The set of words to remove as stopwords.
            store_separate_fields: Whether to store each text attribute as a separate field in the index. This allows for fielded retrieval, but increases index size.
            store_positions: Whether to store position information (i.e., blocks) in the index. This allows for positional queries, but increases index size and retrieval time.
            threads: The number of threads to use during indexing.

        Example Pipeline:

        .. schematic::
            :show_code:

            index = pt.terrier.TerrierIndex('my_index.terrier')
            index.indexer()
        """
        return pt.terrier.IterDictIndexer(
            os.path.realpath(str(self.path)),
            meta=meta,
            text_attrs=text_attrs,
            tokeniser=tokeniser,
            stemmer=stemmer,
            stopwords=stopwords,
            fields=store_separate_fields,
            blocks=store_positions,
            threads=threads,
        )

    def toks_indexer(
        self,
        *,
        meta: Dict = {'docno': 20},
        threads: int = 1,
    ) -> pt.Indexer:
        """Returns an indexer that indexes pre-tokenised documents into this index.

        Args:
            meta: The fields to store as metadata for each document. The keys are the metadata field names, and the values are the maximum lengths for each field.
            threads: The number of threads to use during indexing.

        Example Pipeline:

        .. schematic::
            :show_code:

            index = pt.terrier.TerrierIndex('my_index.terrier')
            index.toks_indexer()
        """
        return pt.terrier.IterDictIndexer(
            os.path.realpath(str(self.path)),
            tokeniser=pt.terrier.TerrierTokeniser.whitespace,
            stemmer=pt.terrier.TerrierStemmer.none,
            stopwords=pt.terrier.TerrierStopwords.none,
            meta=meta,
            pretokenised=True,
            threads=threads,
        )

    def index(self, iter: pt.model.IterDict, **kwargs: Any):
        """Indexes the given input data, creating the index if it does not yet exist, or raising an error if it does.

        This method inspects the first document in the to try to infer reasonable settings for indexing:

        - If a ``toks`` column is present, it is assumed to contain pre-tokenized text, and uses :meth:`toks_indexer`. Otherwise, all ``str`` columns (except ``docno``) are indexed as raw text using :meth:`indexer`.
        - Text fields are stored as metadata.
        - The maximum metadata lengths are predicted based on the lengths present in the first document.

        Args:
            iter: The documents to index as an iterable of dicts.
            kwargs: Ignored
        """
        assert len(kwargs) == 0, f"unknown keyword argument(s) given: {kwargs}"
        assert not self.built(), "an index is already built at this path"

        # TODO: I think we should consider moving some of this logic into IterDictIndexer. It would be really nice to have less configuration to do when indexing.
        iter = pt.utils.peekable(iter)
        first_doc = iter.peek()
        assert 'docno' in first_doc, "input documents must contain a 'docno' field"

        pred_docno_max_length = max(len(first_doc['docno']) * 10, 20) # can we be smarter about this? Perhaps we should sample the first k documents?
        meta = {'docno': pred_docno_max_length}

        text_fields = [k for k, v in sorted(first_doc.items()) if isinstance(v, str) and k != 'docno']
        meta.update({
            field: max(len(first_doc[field]) * 10, 100) # can we be smarter about this? Perhaps we should sample the first k documents?
            for field in text_fields
        })

        if 'toks' in first_doc:
            indexer = self.toks_indexer(meta=meta)
        else:
            assert len(text_fields) >= 1, f"no str or toks fields found in document. Fields: { {k: type(v) for k, v in first_doc.items()} }"
            indexer = self.indexer(text_attrs=text_fields, meta=meta)

        indexer.index(iter)
        return self


    # ----------------------------------------------------
    # Index Data
    # ----------------------------------------------------


    @pt.java.required
    def index_ref(self):
        """The internal Java index reference object for this index.

        Returns:
            A Java `IndexRef <http://terrier.org/docs/current/javadoc/org/terrier/querying/IndexRef.html>`_ object for this index.
        """
        if self._index_ref is None:
            self._index_ref = pt.terrier.J.IndexRef.of(os.path.realpath(str(self.path)))
        return self._index_ref

    def index_obj(self):
        """The internal Java index object for this index.

        Returns:
            A Java `Index <http://terrier.org/docs/current/javadoc/org/terrier/structures/Index.html>`_ object for this index.
        """
        if self._index_obj is None:
            self._index_obj = pt.terrier.IndexFactory.of(self.index_ref(), memory=self.memory)
        return self._index_obj

    def collection_statistics(self):
        """Returns the collection statistics for this index.

        Example:

        .. code-block:: python
            :caption: Show collection statistics for a Terrier index.

            >>> stats = index.collection_statistics()
            >>> print(stats)
            Number of documents: 11429
            Number of terms: 7756
            Number of postings: 224573
            Number of fields: 0
            Number of tokens: 271581
            Field names: []
            Positions:   false

        In this example, the index has 11429 documents, which contained 271581 word occurrences. 7756 unique words were identified. The total number of postings in the inverted index is 224573.
        This index did not record fields during indexing (which can be useful for models such as BM25F). Similarly, positions, which are used for phrasal queries or proximity models were not recorded.

        Returns:
            A Java `CollectionStatistics <http://terrier.org/docs/current/javadoc/org/terrier/structures/CollectionStatistics.html>`_ object for this index.
        """
        return self.index_obj().getCollectionStatistics()

    def lexicon(self):
        """The lexicon for this index.

        Note that the terms in the lexicon include all pre-processing, such as stemming. For example, the term 'chemical' would be
        stored as 'chemic' when using the default Porter stemmer.

        Returns:
            A Java `Lexicon <http://terrier.org/docs/current/javadoc/org/terrier/structures/Lexicon.html>`_ object for this index.
        """
        return self.index_obj().getLexicon()

    def inverted_index(self):
        """The inverted posting index for this index.

        Returns:
            A Java `PostingIndex <http://terrier.org/docs/current/javadoc/org/terrier/structures/PostingIndex.html>`_ object for this index's inverted index.
        """
        return self.index_obj().getInvertedIndex()

    def document_index(self):
        """The document index for this index.

        Returns:
            A Java `DocumentIndex <http://terrier.org/docs/current/javadoc/org/terrier/structures/DocumentIndex.html>`_ object for this index.
        """
        return self.index_obj().getDocumentIndex()

    def meta_index(self):
        """The meta index for this index.

        Example:

        .. code-block:: python
            :caption: Show metadata fields in a Terrier index.

            >>> print(index.meta_index().getKeys())
            ['docno', 'text']

        In this example, the index contains two metadata fields: ``docno``, which contains the document identifiers, and ``text``, which contains the raw text of each document.

        Returns:
            A Java `MetaIndex <http://terrier.org/docs/current/javadoc/org/terrier/structures/MetaIndex.html>`_ object for this index.
        """
        return self.index_obj().getMetaIndex()

    def direct_index(self):
        """The direct (forward) index for this index.

        Returns:
            A Java `PostingIndex <http://terrier.org/docs/current/javadoc/org/terrier/structures/PostingIndex.html>`_ object for this index's direct index.
        """
        return self.index_obj().getDirectIndex()

    # ----------------------------------------------------
    # Miscellaneous
    # ----------------------------------------------------

    def __repr__(self):
        return f'TerrierIndex({str(self.path)!r})'

    def built(self):
        """Returns whether the index has been built (or is a built in-memory index)."""
        return self.path == pt.Artifact.NO_PATH or os.path.exists(os.path.join(self.path, 'data.properties'))

    @classmethod
    @pt.java.required
    def coerce(cls, index_like: Union[str, Path, 'TerrierIndex']) -> 'TerrierIndex':
        """Attempts to build a :class:`TerrierIndex` from the given object.

        Args:
            index_like (object): The object to coerce into a TerrierIndex. If a ``str`` or ``Path``, it loads the index at the provided path. If
                a ``pt.terrier.J.IndexRef`` or ``pt.terrier.J.Index``, it creates a TerrierIndex from the Java object. If a ``pt.terrier.TerrierIndex``,
                it returns itself.
        """
        if isinstance(index_like, TerrierIndex):
            return index_like
        if isinstance(index_like, (str, Path)):
            return TerrierIndex(index_like)
        if isinstance(index_like, pt.terrier.J.IndexRef):
            return TerrierIndex(pt.Artifact.NO_PATH, _index_ref=index_like)
        if isinstance(index_like, pt.terrier.J.Index):
            return TerrierIndex(pt.Artifact.NO_PATH, _index_obj=index_like)
        raise RuntimeError(f'Could not coerce {index_like!r} into a TerrierIndex')

    @staticmethod
    def example():
        """Returns an example Terrier index."""
        return TerrierIndex.from_hf('pyterrier/sample.terrier')

    def get_corpus_iter(self, return_toks: bool = True) -> pt.model.IterDict:
        """Returns an iterable over the documents in this index's corpus.

        Args:
            return_toks: Whether to return tokenised text (list of strings) or raw text (string).

        A corpus iter from a Terrier index can be used for various purposes, including:
          - indexing the pre-tokenised Terrier index directly in another indexing pipeline
          - extracting document metadata for ingestion into another indexing pipeline
        """
        return self.index_obj().get_corpus_iter(return_toks=return_toks)


_WMODEL_MAP: Dict[TerrierModel, str] = {
    TerrierModel.bm25: 'BM25',
    TerrierModel.dph: 'DPH',
    TerrierModel.pl2: 'PL2',
    TerrierModel.dirichlet_lm: 'DirichletLM',
    TerrierModel.hiemstra_lm: 'Hiemstra_LM',
    TerrierModel.tf: 'Tf',
    TerrierModel.tf_idf: 'TF_IDF',
}

def _map_wmodel(model: Union[TerrierModel, str]) -> str:
    if isinstance(model, str):
        return model
    return _WMODEL_MAP[model]

_CONTROL_MAP: Dict[str, str] = {
    'bm25.k1': 'bm25.k_1',
    'bm25.b': 'bm25.b',
    'dfr.c': 'dfr.c',
    'tf_idf.k_1': 'tf_idf.k_1',
    'tf_idf.b': 'tf_idf.b',
    'dirichletlm.mu': 'dirichletlm.mu',
    'hiemstra_lm.lambda': 'hiemstra_lm.lambda',
}
def _map_controls(model_args):
    return {
        _CONTROL_MAP[k]: v
        for k, v in (model_args or {}).items()
        if k in _CONTROL_MAP
    }


_PROPERTY_MAP: Dict[str, str] = {}
def _map_properties(model_args):
    return {
        _PROPERTY_MAP[k]: v
        for k, v in (model_args or {}).items()
        if k in _PROPERTY_MAP
    }



import pandas as pd
from warnings import warn
from typing import List, Union, Callable
from types import FunctionType
import pyterrier as pt
from pyterrier.terrier.index import TerrierTokeniser

@pt.java.required
def tokenise(tokeniser : Union[str,TerrierTokeniser,FunctionType] = 'english', matchop=False) -> pt.Transformer:
    """

    Applies tokenisation to the query. By default, queries obtained from ``pt.get_dataset().get_topics()`` are
    normally tokenised.

    Args:
        tokeniser(Union[str,TerrierTokeniser,FunctionType]): Defines what tokeniser should be used - either a Java tokeniser name in Terrier, a TerrierTokeniser instance, or a function that takes a str as input and returns a list of str.
        matchop(bool): Whether query terms should be wrapped in matchops, to ensure they can be parsed by a Terrier Retriever transformer.
    
    Example - use default tokeniser::

        pipe = pt.rewrite.tokenise() >> pt.terrier.Retriever()
        pipe.search("Question with 'capitals' and other stuff?")
    
    Example - roll your own tokeniser::

        poortokenisation = pt.rewrite.tokenise(lambda query: query.split(" ")) >> pt.terrier.Retriever()

    Example - for non-English languages, tokenise on standard UTF non-alphanumeric characters::

        utftokenised = pt.rewrite.tokenise(pt.TerrierTokeniser.utf)) >> pt.terrier.Retriever()
        utftokenised = pt.rewrite.tokenise("utf")) >> pt.terrier.Retriever()

    Example - tokenising queries using a `HuggingFace tokenizer <https://huggingface.co/docs/transformers/fast_tokenizers>`_ ::

        # this assumes the index was created in a pretokenised manner
        br = pt.terrier.Retriever(indexref)
        tok = AutoTokenizer.from_pretrained("bert-base-uncased")
        query_toks = pt.rewrite.tokenise(tok.tokenize, matchop=True)
        retr_pipe = query_toks >> br
    
    """
    _query_fn: Callable[[str], List[str]]
    if callable(tokeniser):
        _query_fn = tokeniser
    else:
        tokeniser = TerrierTokeniser._to_obj(tokeniser)
        tokeniser = TerrierTokeniser._to_class(tokeniser)
        if "." not in tokeniser:
            tokeniser = 'org.terrier.indexing.tokenisation.' + tokeniser
        tokenobj = pt.java.autoclass(tokeniser)()
        _query_fn = tokenobj.getTokens

    def _join_str(input : Union[str,List[str]]):
        if isinstance(input, str):
            return input
        return ' '.join(input)

    def _join_str_matchop(input : List[str]):
        assert not isinstance(input, str), "Expected a list of strings"
        return ' '.join(map(pt.terrier.Retriever.matchop, input))

    if matchop:
        return pt.apply.query(lambda r: _join_str_matchop(_query_fn(r.query)))
    return pt.apply.query(lambda r: _join_str(_query_fn(r.query)))


class ResetQuery(pt.Transformer):
    def transform(self, inp):
        return pt.model.pop_queries(inp)


def reset() -> pt.Transformer:
    """
        Undoes a previous query rewriting operation. This results in the query formulation stored in the `"query_0"`
        attribute being moved to the `"query"` attribute, and, if present, the `"query_1"` being moved to
        `"query_0"` and so on. This transformation is useful if you have rewritten the query for the purposes
        of one retrieval stage, but wish a subquent transformer to be applies on the original formulation.

        Internally, this function applies `pt.model.pop_queries()`.

        Example::

            firststage = pt.rewrite.SDM() >> pt.terrier.Retriever(index, wmodel="DPH")
            secondstage = pyterrier_bert.cedr.CEDRPipeline()
            fullranker = firststage >> pt.rewrite.reset() >> secondstage

    """
    return ResetQuery()

@pt.java.required
class SDM(pt.Transformer):
    '''
        Implements the sequential dependence model, which Terrier supports using its
        Indri/Galagoo compatible matchop query language. The rewritten query is derived using
        the Terrier class DependenceModelPreProcess. 

        This transformer changes the query. It must be followed by a Terrier Retrieve() transformer.
        The original query is saved in the `"query_0"` column, which can be restored using `pt.rewrite.reset()`.

        .. cite.dblp:: conf/sigir/MetzlerC05
    '''

    def __init__(self, verbose = 0, remove_stopwords = True, prox_model = None, **kwargs):
        super().__init__(**kwargs)
        self.verbose = 0
        self.prox_model = prox_model
        self.remove_stopwords = remove_stopwords
        assert pt.terrier.check_version("5.3")
        self.ApplyTermPipeline_stopsonly = pt.terrier.J.ApplyTermPipeline('Stopwords')

    schematic = {'label': 'SDM'}

    def __repr__(self):
        return "SDM()"

    def transform(self, topics_and_res):
        results = []
        queries = pt.model.ranked_documents_to_queries(topics_and_res)

        # instantiate the DependenceModelPreProcess, specifying a proximity model if specified
        sdm = pt.terrier.J.DependenceModelPreProcess() if self.prox_model is None else pt.terrier.J.DependenceModelPreProcess(self.prox_model)
        
        for row in pt.tqdm(queries.itertuples(), desc=self.name, total=queries.shape[0], unit="q") if self.verbose else queries.itertuples():
            qid = row.qid
            query = row.query
            # parse the querying into a MQT
            rq = pt.terrier.J.Request()
            rq.setQueryID(qid)
            rq.setOriginalQuery(query)
            pt.terrier.J.TerrierQLParser().process(None, rq)
            pt.terrier.J.TerrierQLToMatchingQueryTerms().process(None, rq)
            if self.remove_stopwords:
                self.ApplyTermPipeline_stopsonly.process(None, rq)

            # rewrite the query
            sdm.expandQuery(rq.getMatchingQueryTerms(), rq)
            new_query = ""

            # put the query back into a matchopql form that Terrier can parse later 
            for me in rq.getMatchingQueryTerms():
                term = me.getKey().toString()
                w = me.getValue().getWeight()
                prefix = ""
                if w != 1.0 or me.getValue().termModels.size() > 0:
                    prefix="#combine"
                    if w != 1:
                        prefix += ":0=" + str(w)
                    if me.getValue().termModels.size() == 1:
                        prefix += ":wmodel=" + me.getValue().termModels[0].getClass().getName()
                    term = prefix + "(" + term + ")"
                new_query += term + " "
            new_query = new_query[:-1]
            results.append([qid, new_query])
        new_queries = pd.DataFrame(results, columns=["qid", "query"])
        # restore any other columns, e.g. put back docs if we are re-ranking
        return new_queries.merge(pt.model.push_queries(topics_and_res, inplace=True) , on="qid")

class SequentialDependence(SDM):
    '''
        Implements the sequential dependence model, which Terrier supports using its
        Indri/Galagoo compatible matchop query language. The rewritten query is derived using
        the Terrier class DependenceModelPreProcess. 

        This transformer changes the query. It must be followed by a Terrier Retrieve() transformer.
        The original query is saved in the `"query_0"` column, which can be restored using `pt.rewrite.reset()`.
    '''
    pass

@pt.java.required
class QueryExpansion(pt.Transformer):
    '''
        A base class for applying different types of query expansion using Terrier's classes.
        This transformer changes the query. It must be followed by a Terrier Retrieve() transformer.
        The original query is saved in the `"query_0"` column, which can be restored using `pt.rewrite.reset()`.

        Instance Attributes:
         - fb_terms(int): number of feedback terms. Defaults to 10
         - fb_docs(int): number of feedback documents. Defaults to 3
         
    '''

    def __init__(self, index_like, fb_terms=10, fb_docs=3, qeclass="org.terrier.querying.QueryExpansion", verbose=0, properties={}, requires_scores=False, **kwargs):
        super().__init__(**kwargs)
        self.verbose = verbose
        if isinstance(qeclass, str):
            self.qe = pt.java.autoclass(qeclass)()
        else:
            self.qe = qeclass
        self.indexref = pt.terrier.retriever._parse_index_like(index_like)
        self.properties = properties
        for k,v in properties.items():
            pt.terrier.J.ApplicationSetup.setProperty(k, str(v))
        self.applytp = pt.terrier.J.ApplyTermPipeline()
        self.fb_terms = fb_terms
        self.fb_docs = fb_docs
        self.manager = pt.terrier.J.ManagerFactory._from_(self.indexref)
        self.requires_scores = requires_scores

    def compile(self) -> pt.Transformer:
        return pt.RankCutoff(self.fb_docs) >> self

    def __reduce__(self):
        return (
            self.__class__,
            (self.indexref,),
            self.__getstate__()
        )

    def __getstate__(self): 
        if isinstance(self.qe, str):
            qe = self.qe
        else:
            qe = self.qe.getClass().getName()
        return  {
                'fb_terms' : self.fb_terms, 
                'fb_docs' : self.fb_docs,
                'qeclass' : qe,
                'properties' : self.properties
                }

    def __setstate__(self, d): 
        self.fb_terms = d["fb_terms"]
        self.fb_docs = d["fb_docs"]
        self.qe = pt.java.autoclass(d['qeclass'])()
        self.properties.update(d["properties"])
        for key,value in d["properties"].items():
            self.appSetup.setProperty(key, str(value))
        self.manager = pt.terrier.J.ManagerFactory._from_(self.indexref)

    def _populate_resultset(self, topics_and_res, qid, index):
        
        docids = None
        scores = None
        occurrences = None
        if "docid" in topics_and_res.columns:
            # we need .tolist() as jnius cannot convert numpy arrays
            topics_and_res_for_qid = topics_and_res[topics_and_res["qid"] == qid]
            docids = topics_and_res_for_qid["docid"].values.tolist()
            scores = [0.0] * len(docids)
            if self.requires_scores:
                scores = topics_and_res_for_qid["score"].values.tolist()
            occurrences = [0] * len(docids)

        elif "docno" in topics_and_res.columns:
            topics_and_res_for_qid = topics_and_res[topics_and_res["qid"] == qid]
            docnos = topics_and_res_for_qid["docno"].values
            docids = []
            scores = []
            _scores = [0.0] * len(docids)
            if self.requires_scores:
                _scores = topics_and_res_for_qid["score"].values.tolist()

            occurrences = []
            metaindex = index.getMetaIndex()
            skipped = 0
            for docno, docscore in zip(docnos, _scores):
                docid = metaindex.getDocument("docno", docno)
                if docid == -1:
                    skipped +=1 
                assert docid != -1, "could not match docno" + docno + " to a docid for query " + qid    
                docids.append(docid)
                occurrences.append(0)
                scores.append(docscore)
            if skipped > 0:
                if skipped == len(docnos):
                    warn(
                        "*ALL* %d feedback docnos for qid %s could not be found in the index" % (skipped, qid))
                else:
                    warn(
                        "%d feedback docnos for qid %s could not be found in the index" % (skipped, qid))
        else:
            raise ValueError("Input resultset has neither docid nor docno")
        return pt.terrier.J.QueryResultSet(docids, scores, occurrences)

    def __repr__(self):
        return "QueryExpansion(" + ",".join([
            self.indexref.toString(),
            str(self.fb_docs),
            str(self.fb_terms),
            str(self.qe)
            ]) + ")"

    def _configure_request(self, rq):
        rq.setControl("qe_fb_docs", str(self.fb_docs))
        rq.setControl("qe_fb_terms", str(self.fb_terms))

    def transform(self, topics_and_res):
        with pt.validate.any(topics_and_res) as v:
            v.columns(includes=['qid', 'docno', 'query'])
            v.columns(includes=['qid', 'docid', 'query'])

        results = []

        queries = pt.model.ranked_documents_to_queries(topics_and_res)
        #queries = topics_and_res[query_columns(topics_and_res, qid=True)].dropna(axis=0, subset=query_columns(topics_and_res, qid=False)).drop_duplicates()
                
        for row in pt.tqdm(queries.itertuples(), desc=self.name, total=queries.shape[0], unit="q") if self.verbose else queries.itertuples():
            qid = row.qid
            query = row.query
            srq = self.manager.newSearchRequest(qid, query)
            rq = pt.java.cast("org.terrier.querying.Request", srq)
            self.qe.configureIndex(rq.getIndex())
            self._configure_request(rq)

            # generate the result set from the input
            rq.setResultSet(self._populate_resultset(topics_and_res, qid, rq.getIndex()))

            
            pt.terrier.J.TerrierQLParser().process(None, rq)
            pt.terrier.J.TerrierQLToMatchingQueryTerms().process(None, rq)
            # how to make sure this happens/doesnt happen when appropriate.
            self.applytp.process(None, rq)
            # to ensure weights are identical to Terrier
            rq.getMatchingQueryTerms().normaliseTermWeights()
            self.qe.expandQuery(rq.getMatchingQueryTerms(), rq)

            # this control for Terrier stops it re-stemming the expanded terms
            new_query = "applypipeline:off "
            for me in rq.getMatchingQueryTerms():
                new_query += me.getKey().toString() + ( "^%.9f "  % me.getValue().getWeight() ) 
            # remove trailing space
            new_query = new_query[:-1]
            results.append([qid, new_query])
        new_queries = pd.DataFrame(results, columns=["qid", "query"])
        return pt.model.push_queries(queries, inplace=True).merge(new_queries, on="qid")

class DFRQueryExpansion(QueryExpansion):

    def __init__(self, *args, qemodel="Bo1", **kwargs):
        super().__init__(*args, **kwargs)
        self.qemodel = qemodel

    def _configure_request(self, rq):
        super()._configure_request(rq)
        rq.setControl("qemodel", self.qemodel)

class Bo1QueryExpansion(DFRQueryExpansion):
    '''
        Applies the Bo1 query expansion model from the Divergence from Randomness Framework, as provided by Terrier.
        It must be followed by a Terrier Retrieve() transformer.
        The original query is saved in the `"query_0"` column, which can be restored using `pt.rewrite.reset()`.

        Instance Attributes:
         - fb_terms(int): number of feedback terms. Defaults to 10
         - fb_docs(int): number of feedback documents. Defaults to 3  

        .. cite.dblp::     journals/tois/AmatiR02
    '''
    schematic = {'label': 'Bo1'}

    def __init__(self, *args, **kwargs):
        """
        Args:
            index_like: the Terrier index to use.
            fb_terms(int): number of terms to add to the query. Terrier's default setting is 10 expansion terms.
            fb_docs(int): number of feedback documents to consider. Terrier's default setting is 3 feedback documents.
        """
        kwargs["qemodel"] = "Bo1"
        super().__init__(*args, **kwargs)

class KLQueryExpansion(DFRQueryExpansion):
    '''
        Applies the KL query expansion model from the Divergence from Randomness Framework, as provided by Terrier.
        This transformer must be followed by a Terrier Retrieve() transformer.
        The original query is saved in the `"query_0"` column, which can be restored using `pt.rewrite.reset()`.

        Instance Attributes:
         - fb_terms(int): number of feedback terms. Defaults to 10
         - fb_docs(int): number of feedback documents. Defaults to 3  
    '''
    schematic = {'label': 'KL'}

    def __init__(self, *args, **kwargs):
        """
        Args:
            index_like: the Terrier index to use
            fb_terms(int): number of terms to add to the query. Terrier's default setting is 10 expansion terms.
            fb_docs(int): number of feedback documents to consider. Terrier's default setting is 3 feedback documents.
        """
        kwargs["qemodel"] = "KL"
        super().__init__(*args, **kwargs)

@pt.java.required
class RM3(QueryExpansion):
    '''
        Performs query expansion using RM3 relevance models.

        This transformer must be followed by a Terrier Retrieve() transformer.
        The original query is saved in the `"query_0"` column, which can be restored using `pt.rewrite.reset()`.

        Instance Attributes:
         - fb_terms(int): number of feedback terms. Defaults to 10
         - fb_docs(int): number of feedback documents. Defaults to 3
         - fb_lambda(float): lambda in RM3, i.e. importance of relevance model viz feedback model. Defaults to 0.6.

        Example:: 

            bm25 = pt.terrier.Retriever(index, wmodel="BM25")
            rm3_pipe = bm25 >> pt.rewrite.RM3(index) >> bm25
            pt.Experiment([bm25, rm3_pipe],
                        dataset.get_topics(),
                        dataset.get_qrels(),
                        ["map"]
                        )

        .. cite.dblp::     conf/trec/JaleelACDLLSW04
 
    '''
    def __init__(self, *args, fb_terms=10, fb_docs=3, fb_lambda=0.6, **kwargs):
        """
        Args:
            index_like: the Terrier index to use
            fb_terms(int): number of terms to add to the query. Terrier's default setting is 10 expansion terms.
            fb_docs(int): number of feedback documents to consider. Terrier's default setting is 3 feedback documents.
            fb_lambda(float): lambda in RM3, i.e. importance of relevance model viz feedback model. Defaults to 0.6.
        """
        assert pt.terrier.check_version("5.10"), "Terrier 5.10 required"
        rm = pt.terrier.J.RM3()
        self.fb_lambda = fb_lambda
        kwargs["qeclass"] = rm
        super().__init__(*args, fb_terms=fb_terms, fb_docs=fb_docs, requires_scores=True, **kwargs)

    def __getstate__(self): 
        rtr = super().__getstate__()
        rtr['fb_lambda'] = self.fb_lambda
        return rtr
    
    def __setstate__(self, d): 
        super().__setstate__(d)
        self.fb_lambda = d["fb_lambda"]

    def _configure_request(self, rq):
        super()._configure_request(rq)
        rq.setControl("rm3.lambda", str(self.fb_lambda))
        
    def transform(self, queries_and_docs):
        self.qe.fbTerms = self.fb_terms
        self.qe.fbDocs = self.fb_docs
        return super().transform(queries_and_docs)

def stash_results(clear=True) -> pt.Transformer:
    """
    Stashes (saves) the current retrieved documents for each query into the column `"stashed_results_0"`.
    This means that they can be restored later by using `pt.rewrite.reset_results()`.
    thereby converting a retrieved documents dataframe into one of queries.

    Args: 
    clear(bool): whether to drop the document and retrieved document related columns. Defaults to True.

    """
    return _StashResults(clear)
    
def reset_results() -> pt.Transformer:
    """
    Applies a transformer that undoes a `pt.rewrite.stash_results()` transformer, thereby restoring the
    ranked documents.
    """
    return _ResetResults()

class _StashResults(pt.Transformer):

    def __init__(self, clear, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.clear = clear

    def transform(self, topics_and_res: pd.DataFrame) -> pd.DataFrame:
        if "stashed_results_0" in topics_and_res.columns:
            raise ValueError("Cannot apply pt.rewrite.stash_results() more than once")
        doc_cols = pt.model.document_columns(topics_and_res)
        
        rtr =  []
        if self.clear:
            query_cols = pt.model.query_columns(topics_and_res)            
            for qid, groupDf in topics_and_res.groupby("qid"):
                documentsDF = groupDf[doc_cols]
                queryDf = groupDf[query_cols].iloc[0]
                queryDict = queryDf.to_dict()
                queryDict["stashed_results_0"] = documentsDF.to_dict(orient='records')
                rtr.append(queryDict)
            return pd.DataFrame(rtr)
        else:
            for qid, groupDf in topics_and_res.groupby("qid"):
                groupDf = groupDf.reset_index().copy()
                documentsDF = groupDf[doc_cols]
                docsDict = documentsDF.to_dict(orient='records')
                groupDf["stashed_results_0"] = None
                for i in range(len(groupDf)):
                    groupDf.at[i, "stashed_results_0"] = docsDict
                rtr.append(groupDf)
            return pd.concat(rtr)   

    def __repr__(self):
        return "pt.rewrite.stash_results()"     

class _ResetResults(pt.Transformer):

    def transform(self, topics_with_saved_docs : pd.DataFrame) -> pd.DataFrame:
        if "stashed_results_0" not in topics_with_saved_docs.columns:
            raise ValueError("Cannot apply pt.rewrite.reset_results() without pt.rewrite.stash_results() - column stashed_results_0 not found")
        rtr = []
        for row in topics_with_saved_docs.itertuples():
            docsdf = pd.DataFrame.from_records(row.stashed_results_0)
            docsdf["qid"] = row.qid
            querydf = pd.DataFrame(data=[row])
            querydf.drop("stashed_results_0", axis=1, inplace=True)
            finaldf = querydf.merge(docsdf, on="qid")
            rtr.append(finaldf)
        return pd.concat(rtr)

    def __repr__(self):
        return "pt.rewrite.reset_results()"

def linear(weightCurrent : float, weightPrevious : float, format="terrierql", **kwargs) -> pt.Transformer:
    """
    Applied to make a linear combination of the current and previous query formulation. The implementation
    is tied to the underlying query language used by the retrieval/re-ranker transformers. Two of Terrier's
    query language formats are supported by the `format` kwarg, namely `"terrierql"` and `"matchoptql"`. 
    Their exact respective formats are `detailed in the Terrier documentation <https://github.com/terrier-org/terrier-core/blob/5.x/doc/querylanguage.md>`_.

    Args:
        weightCurrent(float): weight to apply to the current query formulation.
        weightPrevious(float): weight to apply to the previous query formulation.
        format(str): which query language to use to rewrite the queries, one of "terrierql" or "matchopql".

    Example::

        pipeTQL = pt.apply.query(lambda row: "az") >> pt.rewrite.linear(0.75, 0.25, format="terrierql")
        pipeMQL = pt.apply.query(lambda row: "az") >> pt.rewrite.linear(0.75, 0.25, format="matchopql")
        pipeT.search("a")
        pipeM.search("a")

    Example outputs of `pipeTQL` and `pipeMQL` corresponding to the query "a" above:

    - Terrier QL output: `"(az)^0.750000 (a)^0.250000"`
    - MatchOp QL output: `"#combine:0:0.750000:1:0.250000(#combine(az) #combine(a))"`

    """
    return _LinearRewriteMix([weightCurrent, weightPrevious], format, **kwargs)

class _LinearRewriteMix(pt.Transformer):

    def __init__(self, weights : List[float], format : str = 'terrierql', **kwargs):
        super().__init__(**kwargs)
        self.weights = weights
        self.format = format
        if format not in ["terrierql", "matchopql"]:
            raise ValueError("Format must be one of 'terrierql', 'matchopql'")

    def _terrierql(self, row):
        return "(%s)^%f (%s)^%f" % (
            row["query_0"],
            self.weights[0],
            row["query_1"],
            self.weights[1])
    
    def _matchopql(self, row):
        return "#combine:0:%f:1:%f(#combine(%s) #combine(%s))" % (
            self.weights[0],
            self.weights[1],
            row["query_0"],
            row["query_1"])

    def transform(self, topics_and_res):
        fn = None
        if self.format == "terrierql":
            fn = self._terrierql
        elif self.format == "matchopql":
            fn = self._matchopql

        newDF = pt.model.push_queries(topics_and_res)
        newDF["query"] = newDF.apply(fn, axis=1)
        return newDF

    def __repr__(self):
        return "pt.rewrite.linear()"



import pandas as pd
import unittest
import os
import pyterrier as pt
from .base import BaseTestCase, ensure_deprecated
import warnings

class TestFeaturesBatchRetrieve(BaseTestCase):

    def test_compile_funion_to_fbr(self):

        indexdir = self.here + "/fixtures/index/data.properties"
        indexref = pt.IndexRef.of(indexdir)
        index = pt.IndexFactory.of(indexref)
        input = pd.DataFrame([["1", "Stability"]], columns=['qid', 'query'])

        for indexloc in [indexdir, indexref, index]:
            # we only want a candidate set of 2 documents
            firstpass = pt.terrier.Retriever(indexloc, wmodel="BM25")
            pipe_f_fbr = firstpass >> pt.terrier.FeaturesRetriever(indexloc, features=["WMODEL:DPH", "WMODEL:PL2"])
            pipe_fbr = pt.terrier.FeaturesRetriever(indexloc, wmodel="BM25", features=["WMODEL:DPH", "WMODEL:PL2"])
            pipe_raw = firstpass >> ( pt.terrier.Retriever(indexloc, wmodel="DPH") ** pt.terrier.Retriever(indexloc, wmodel="PL2") )

            self.assertEqual(pipe_raw[1][0].indexref, firstpass.indexref, "indexref %s not the same as indexref %s" % (pipe_raw[1][0].indexref.toString(), firstpass.indexref.toString()))
            self.assertEqual(pipe_raw[1][0].indexref, pipe_raw[1][1].indexref, "indexref %s not the same as indexref %s" % (pipe_raw[1][0].indexref.toString(), pipe_raw[1][1].indexref.toString()))
            
            res1 = (pipe_f_fbr %2)(input)[['qid', 'query', 'docid', 'rank', 'features', 'docno', 'score']]
            res2 = (pipe_fbr % 2)(input)[['qid', 'query', 'docid', 'rank', 'features', 'docno', 'score']]
            pd.testing.assert_frame_equal(res1, res2)
            res3 = (pipe_raw % 2)(input)[['qid', 'query', 'docid', 'rank', 'features', 'docno', 'score']]
            pd.testing.assert_frame_equal(res1, res3)

            compiled = (pipe_raw % 2).compile()
            self.assertTrue( isinstance(compiled, pt.terrier.FeaturesRetriever) , 
                            "compiled is not a FeaturesRetriever, it was %s for indexloc %s" % (str(compiled), indexloc))
            self.assertEqual("BM25", compiled.wmodel)
            print(repr(compiled))
            self.assertEqual("1", compiled.controls["end"])
            res4 = compiled(input)[['qid', 'query', 'docid', 'rank', 'features', 'docno', 'score']] #reorder dataframe columns
            res1 = res1[['qid', 'query', 'docid', 'rank', 'features', 'docno', 'score']]
            pd.testing.assert_frame_equal(res1, res4)
        

    def test_compile_left_to_fbr(self):
        indexdir = self.here + "/fixtures/index/data.properties"
        indexref = pt.IndexRef.of(indexdir)
        index = pt.IndexFactory.of(indexref)
        for indexloc in [indexdir, indexref, index]:
            # we only want a candidate set of 2 documents
            firstpass = pt.terrier.Retriever(indexloc, wmodel="BM25")
            pipe_f_fbr = firstpass >> pt.terrier.FeaturesRetriever(indexloc, features=["WMODEL:DPH", "WMODEL:PL2"])
            pipe_fbr = pt.terrier.FeaturesRetriever(indexloc, wmodel="BM25", features=["WMODEL:DPH", "WMODEL:PL2"])
            input = pd.DataFrame([["1", "Stability"]], columns=['qid', 'query'])
            res1 = (pipe_fbr %2)(input)
            
            compiled = (pipe_f_fbr % 2).compile()
            self.assertTrue( isinstance(compiled, pt.terrier.FeaturesRetriever) )
            self.assertEqual("BM25", compiled.wmodel)
            self.assertTrue("1", compiled.controls["end"])
            res4 = compiled(input)
            pd.testing.assert_frame_equal(res1, res4)

    def test_fbr_reranking(self):
        if not pt.terrier.check_version("5.4"):
            self.skipTest("Requires Terrier 5.4")
        # this test examines the use of ScoringMatchingWithFat 
        JIR = pt.java.autoclass('org.terrier.querying.IndexRef')
        indexref = JIR.of(self.here + "/fixtures/index/data.properties")
        # we only want a candidate set of 2 documents
        firstpass = pt.terrier.Retriever(indexref, wmodel="BM25") % 2
        pipe = firstpass >> pt.terrier.FeaturesRetriever(indexref, features=["WMODEL:DPH", "WMODEL:PL2"])
        input = pd.DataFrame([["1", "Stability"]], columns=['qid', 'query'])

        result0 = firstpass(input)
        result0_map = { row.docno : row.score for row in result0.itertuples() }

        result = pipe.transform(input)
        self.assertTrue("qid" in result.columns)
        self.assertTrue("docno" in result.columns)
        self.assertTrue("score" in result.columns)
        self.assertTrue("features" in result.columns)
        self.assertEqual(2, len(result))
        self.assertEqual(result.iloc[0]["features"].size, 2)
        
        result1S_map = { row.docno : row.score for row in result.itertuples() }
        self.assertEqual(result0_map, result1S_map)

        pipe_simple = firstpass >> (pt.terrier.Retriever(indexref, wmodel="DPH") ** pt.terrier.Retriever(indexref, wmodel="PL2"))
        result2 = pipe.transform(input)
        import numpy as np
        f1 = np.stack(result["features"].values)
        f2 = np.stack(result2["features"].values)
        self.assertTrue( np.array_equal(f1, f2))

        result2S_map = { row.docno : row.score for row in result2.itertuples() }
        self.assertEqual(result0_map, result2S_map)

        result1F0_map = { row.docno : row.features[0] for row in result.itertuples() }
        result2F0_map = { row.docno : row.features[0] for row in result2.itertuples() }
        result1F1_map = { row.docno : row.features[1] for row in result.itertuples() }
        result2F1_map = { row.docno : row.features[1] for row in result2.itertuples() }

        self.assertEqual(result1F0_map, result2F0_map)
        self.assertEqual(result1F1_map, result2F1_map)

    def test_fbr_reranking2(self):
        if not pt.terrier.check_version("5.4"):
            self.skipTest("Requires Terrier 5.4")
        # this test examines the use of ScoringMatchingWithFat, using a particular case known to with Terrier 5.3 
        JIR = pt.java.autoclass('org.terrier.querying.IndexRef')
        indexref = JIR.of(self.here + "/fixtures/index/data.properties")
        # we only want a candidate set of 3 documents
        firstpass = pt.terrier.Retriever(indexref, wmodel="BM25") % 3
        pipe1 = firstpass >> pt.terrier.FeaturesRetriever(indexref, features=["WMODEL:PL2"])
        pipe2 = firstpass >> pt.terrier.Retriever(indexref, wmodel="PL2")
        
        input = pd.DataFrame([["1", "Stability"]], columns=['qid', 'query'])
        result0 = firstpass.transform(input)
        result1 = pipe1.transform(input)
        result2 = pipe2.transform(input)

        result1["feature0"] = result1.apply(lambda row: row["features"][0], axis=1)
        #BM25 score
        result0_map = { row.docno : row.score for row in result0.itertuples() }
        result1S_map = { row.docno : row.score for row in result1.itertuples() }
        #PL2 score
        result1F_map = { row.docno : row.feature0 for row in result1.itertuples() }
        result2_map = { row.docno : row.score for row in result2.itertuples() }

        # check features scores
        # NB: places can go no less than 4, as two documents have similar PL2 scores
        for rank, row in enumerate(result0.itertuples()):
            docno = row.docno
            # check that score is unchanged
            self.assertAlmostEqual(result1S_map[docno], result0_map[docno], msg="input score mismatch at rank %d for docno %s" % (rank, docno), places=4)
            #  check that feature score is correct
            self.assertAlmostEqual(result1F_map[docno], result2_map[docno], msg="feature score mismatch at rank %d for docno %s" % (rank, docno), places=4)

    def test_fbr_ltr(self):
        JIR = pt.java.autoclass('org.terrier.querying.IndexRef')
        indexref = JIR.of(self.here + "/fixtures/index/data.properties")
        retr = pt.terrier.FeaturesRetriever(indexref, ["WMODEL:PL2"], wmodel="DPH")
        topics = pt.io.read_topics(self.here + "/fixtures/vaswani_npl/query-text.trec").head(3)
        qrels = pt.io.read_qrels(self.here + "/fixtures/vaswani_npl/qrels")
        res = retr.transform(topics)
        res = res.merge(qrels, on=['qid', 'docno'], how='left').fillna(0)
        from sklearn.ensemble import RandomForestClassifier
        import numpy as np
        #print(res.dtypes)
        RandomForestClassifier(n_estimators=10).fit(np.stack(res["features"]), res["label"])

    def test_fbr(self):
        JIR = pt.java.autoclass('org.terrier.querying.IndexRef')
        indexref = JIR.of(self.here + "/fixtures/index/data.properties")

        
        def do_test(name, retr):
            retr = retr(indexref, ["WMODEL:PL2"], wmodel="DPH")
            self.subTest(name)
            input = pd.DataFrame([["1", "Stability"]], columns=['qid', 'query'])
            result = retr.transform(input)
            self.assertTrue("qid" in result.columns)
            self.assertTrue("docno" in result.columns)
            self.assertTrue("score" in result.columns)
            self.assertTrue("features" in result.columns)
            self.assertTrue(len(result) > 0)
            self.assertEqual(result.iloc[0]["features"].size, 1)

            input = pd.DataFrame([["1", "Stability", "u1"]], columns=['qid', 'query', 'username'])
            result = retr.transform(input)
            self.assertIn("username", result.columns)

            retrBasic = pt.terrier.Retriever(indexref)
            if "matching" in retrBasic.controls:
                self.assertNotEqual(retrBasic.controls["matching"], "FatFeaturedScoringMatching,org.terrier.matching.daat.FatFull")

        @ensure_deprecated 
        def do_test_deprecated(name, retr):
            do_test(name, retr)

        do_test("FeaturesRetriever", pt.terrier.FeaturesRetriever)
        do_test_deprecated("FeaturesBatchRetrieve", pt.FeaturesBatchRetrieve)
    
    def test_fbr_query_toks(self):
        indexloc = self.here + "/fixtures/index/data.properties"
        
        retr = pt.terrier.FeaturesRetriever(indexloc, ["WMODEL:PL2"], wmodel="DPH")
        query_terrier = 'applypipeline:off chemic^2 reaction^0.5'
        result_terrier = retr.search(query_terrier)

        query_matchop = '#combine:0=2:1=0.5(chemic reaction)'
        result_matchop = retr.search(query_matchop)

        query_toks = { 'chemic' : 2, 'reaction' : 0.5}
        result_toks = retr.transform(pd.DataFrame([['1', query_toks]], columns=['qid', 'query_toks']))
        
        self.assertEqual(len(result_terrier), len(result_matchop))
        self.assertEqual(len(result_terrier), len(result_toks))
        from pandas.testing import assert_frame_equal
        assert_frame_equal(result_terrier[["qid", "docno", "score", "rank", "features"]], result_matchop[["qid", "docno", "score", "rank", "features"]])
        assert_frame_equal(result_terrier[["qid", "docno", "score", "rank", "features"]], result_toks[["qid", "docno", "score", "rank", "features"]])

    def test_fbr_example(self):
        JIR = pt.java.autoclass('org.terrier.querying.IndexRef')
        indexref = JIR.of(self.here + "/fixtures/index/data.properties")
        index = pt.IndexFactory.of(indexref)
        # this ranker will make the candidate set of documents for each query
        BM25 = pt.terrier.Retriever(index, wmodel="BM25")

        # these rankers we will use to re-rank the BM25 results
        TF_IDF =  pt.terrier.Retriever(index, wmodel="Dl")
        PL2 =  pt.terrier.Retriever(index, wmodel="PL2")

        pipe =  (BM25 %2) >> (TF_IDF ** PL2)
        fbr = pt.terrier.FeaturesRetriever(indexref, ["WMODEL:Dl", "WMODEL:PL2"], wmodel="BM25") % 2
        resultP = pipe.search("chemical")
        resultF = fbr.search("chemical")
        pd.set_option('display.max_columns', None)

        self.assertEqual(resultP.iloc[0].docno, resultF.iloc[0].docno)
        self.assertEqual(resultP.iloc[0].score, resultF.iloc[0].score)
        self.assertEqual(resultP.iloc[0].features[0], resultF.iloc[0].features[0])
        self.assertEqual(resultP.iloc[0].features[1], resultF.iloc[0].features[1])

        pipeCompiled = pipe.compile()
        resultC = pipeCompiled.search("chemical")
        self.assertEqual(resultP.iloc[0].docno, resultC.iloc[0].docno)
        self.assertEqual(resultP.iloc[0].score, resultC.iloc[0].score)
        self.assertEqual(resultP.iloc[0].features[0], resultC.iloc[0].features[0])
        self.assertEqual(resultP.iloc[0].features[1], resultC.iloc[0].features[1])

    def test_fbr_empty(self):
        JIR = pt.java.autoclass('org.terrier.querying.IndexRef')
        indexref = JIR.of(self.here + "/fixtures/index/data.properties")
        retr = pt.terrier.FeaturesRetriever(indexref, ["WMODEL:PL2"], wmodel="DPH")
        input = pd.DataFrame([["1", ""]], columns=['qid', 'query'])
        with warnings.catch_warnings(record=True) as w:
            result = retr.transform(input)
            assert "Skipping empty query" in str(w[-1].message)
        self.assertTrue(len(result) == 0)

if __name__ == "__main__":
    unittest.main()



import pandas as pd
from . import Transformer
from typing import List, Optional

def print_columns(by_query : bool = False, message : Optional[str] = None) -> Transformer:
    """
    Returns a transformer that can be inserted into pipelines that can print the column names of the dataframe
    at this stage in the pipeline:

    :param by_query: whether to display for each query. Defaults to False.
    :param message: whether to display a message before printing. Defaults to None, which means no message. This
       is useful when ``print_columns()`` is being used multiple times within a pipeline 
     

    Example::
    
        pipe = (
            bm25
            >> pt.debug.print_columns() 
            >> pt.rewrite.RM3() 
            >> pt.debug.print_columns()
            bm25)

    When the above pipeline is executed, two sets of columns will be displayed
     - `["qid", "query", "docno", "rank", "score"]`  - the output of BM25, a ranking of documents
     - `["qid", "query", "query_0"]`   - the output of RM3, a reformulated query
    
        
    """
    import pyterrier as pt
    def _do_print(df):
        if message is not None:
            print(message)
        print(df.columns)
        return df
    return pt.apply.by_query(_do_print) if by_query else pt.apply.generic(_do_print) 

def print_num_rows(
        by_query : bool = True, 
        msg : str = "num_rows") -> Transformer:
    """
    Returns a transformer that can be inserted into pipelines that can print the number of rows names of the dataframe
    at this stage in the pipeline:

    :param by_query: whether to display for each query. Defaults to True.
    :param message: whether to display a message before printing. Defaults to "num_rows". This
       is useful when ``print_columns()`` is being used multiple times within a pipeline 
     
    Example::
    
        pipe = (
            bm25
            >> pt.debug.print_num_rows() 
            >> pt.rewrite.RM3() 
            >> pt.debug.print_num_rows()
            bm25)

    When the above pipeline is executed, the following output will be displayed
     - `num_rows 1: 1000` - the output of BM25, a ranking of documents
     - `num_rows 1: 1` - the output of RM3, the reformulated query
    
        
    """

    import pyterrier as pt
    def _print_qid(df):
        qid = df.iloc[0].qid
        print("%s %s: %d" % (msg, qid, len(df)))
        return df

    def _print(df):
        print("%s: %d" % (msg, len(df)))
        return df

    if by_query:
        return pt.apply.by_query(_print_qid, add_ranks=False)
    else:
        return pt.apply.generic(_print, add_ranks=False)

def print_rows(
        by_query : bool = True, 
        jupyter: bool = True, 
        head : int = 2, 
        message : Optional[str] = None, 
        columns : Optional[List[str]] = None) -> Transformer:
    """
    Returns a transformer that can be inserted into pipelines that can print some of the dataframe
    at this stage in the pipeline:

    :param by_query: whether to display for each query. Defaults to True.
    :param jupyter: Whether to use IPython's display function to display the dataframe. Defaults to True.
    :param head: The number of rows to display. None means all rows.
    :param columns: Limit the columns for which data is displayed. Default of None displays all columns.
    :param message: whether to display a message before printing. Defaults to None, which means no message. This
       is useful when ``print_rows()`` is being used multiple times within a pipeline 

    Example::

        pipe = (
            bm25
            >> pt.debug.print_rows() 
            >> pt.rewrite.RM3() 
            >> pt.debug.print_rows()
            bm25)
     
    """
    import pyterrier as pt
    def _do_print(df):
        if message is not None:
            print(message)
        render = df if head is None else df.head(head)
        if columns is not None:
            render = render[columns]
        if jupyter:
            from IPython.display import display # type: ignore
            display(render)
        else:
            print(render)
        return df
    return pt.apply.by_query(_do_print) if by_query else pt.apply.generic(_do_print) 

class pdb(Transformer):
    """Returns a transformer that starts an interactive `pdb <https://docs.python.org/3/library/pdb.html>`__
    debugger session. The interactive session can be used to inspect the dataframe at this stage in the pipeline.

    Example::

        pipe = (
            bm25
            >> pt.debug.pdb()
            >> pt.rewrite.RM3() 
            >> pt.debug.pdb()
            bm25)
    """
    def transform(self, inp: pd.DataFrame) -> pd.DataFrame:
        import pdb
        pdb.set_trace()
        return inp



import pyterrier as pt

import unittest
import tempfile
import shutil
import os
import numpy as np
from .base import BaseTestCase

class TestBackground(BaseTestCase):

    @staticmethod
    def get_contents(docid, index):
        rtr = {}
        direct = index.getDirectIndex()
        lexicon = index.getLexicon()
        for p in direct.getPostings(index.getDocumentIndex().getDocumentEntry(docid)):
            rtr[lexicon.getLexiconEntry(p.getId()).getKey()] = p.getFrequency()
        return rtr
    
    def test2_manualM(self):
        self._test2_manual(pt.index.IndexingType.MEMORY)
    def test2_manualC(self):
        self._test2_manual(pt.index.IndexingType.CLASSIC)
    def test2_manualS(self):
        self._test2_manual(pt.index.IndexingType.SINGLEPASS)
                

    def _test2_manual(self, type):
        import pyterrier as pt
        #pt.logging("INFO")
        import pandas as pd
        df1 = pd.DataFrame({
            'docno': ['1048'],
            'text':
                ['h  f  noise radiators in ground flashes of tropical lightning  a '+
                'detailed analysis of h  f  noise sources in tropical ground flashes '+
                'v  l  f  phase characteristics deduced from atmospheric waveforms']
        })
        pd_indexer1 = pt.IterDictIndexer(tempfile.mkdtemp(), type=type)
        indexref1 = pd_indexer1.index(df1.to_dict(orient='records'))
        index1 = pt.IndexFactory.of(indexref1)
        
        has_direct1 = index1.hasIndexStructure("direct")

        indexref_big = pt.get_dataset("vaswani").get_index()
        index_big = pt.IndexFactory.of(indexref_big)

        stopwords = pt.java.autoclass("org.terrier.terms.Stopwords")(None)
        stemmer = pt.java.autoclass("org.terrier.terms.PorterStemmer")(None)

        q = "MATHEMATICAL ANALYSIS AND DESIGN DETAILS OF WAVEGUIDE FED MICROWAVE RADIATIONS"
        self.assertEqual("1048", index_big.getMetaIndex().getItem("docno", 1047))
        contents_big = TestBackground.get_contents(1047, index_big)

        def _check_index(index_small):
            if has_direct1:
                contents1 = TestBackground.get_contents(0, index_small)
                self.assertEqual(contents1, contents_big)

            inv1 = index_small.getInvertedIndex()
            print(inv1.getClass().getName())
            lex1 = index_small.getLexicon()
            for t in contents_big:
                pointer = lex1[t]
                print(pointer.toString())
                p = inv1.getPostings(pointer)
                print(p.getClass().getName())
                rtr = p.next()
                self.assertEqual(0, rtr)
                self.assertEqual(p.getDocumentLength(), index_big.getDocumentIndex().getDocumentLength(1047))
                self.assertEqual(contents_big[t], p.getFrequency())
                self.assertEqual(p.next(), p.EOL)

            from jnius import JavaException
            try:
                br1 = pt.terrier.Retriever(index_small, wmodel="Tf")
                brall = pt.terrier.Retriever(index_big, wmodel="Tf")
                with_doc = pd.DataFrame([["q1", q, "1048", 1047]], columns=["qid", "query", "docno", "docid"])
                rtr1 = br1.search(q)
            except JavaException as ja:
                print(ja.stacktrace)
                raise ja
            rtrall = brall(with_doc)            
            self.assertTrue(np.array_equal(rtr1["score"].values, rtrall["score"].values))
        
        _check_index(index1)
        _check_index( pt.java.autoclass("org.terrier.python.IndexWithBackground")(index1, index_big))

    def test_itM(self):
        self._test_it(pt.index.IndexingType.MEMORY)
    def test_itC(self):
        self._test_it(pt.index.IndexingType.CLASSIC)
    def test_itS(self):
        self._test_it(pt.index.IndexingType.SINGLEPASS)
        

    def _test_it(self, type):
        import pyterrier as pt
        import pandas as pd
        df1 = pd.DataFrame({
            'docno': ['1', '2', '3'],
            'url':
                ['url1', 'url2', 'url3'],
            'text':
                ['He ran out of money, so he had to stop playing',
                 'The wave were crash on the shore; it was a',
                 'The body may perhaps compensates for the loss']
        })
        pd_indexer1 = pt.IterDictIndexer(tempfile.mkdtemp(), type=type)
        indexref1 = pd_indexer1.index(df1.to_dict(orient='records'))

        df2 = pd.DataFrame({
            'docno': ['14'],
            'text': ['test wave']
        })

        from jnius import JavaException
        try:

            pd_indexer2 = pt.IterDictIndexer(tempfile.mkdtemp(), type=type)
            indexref2 = pd_indexer2.index(df2.to_dict(orient='records'))
            
            index1 = pt.IndexFactory.of(indexref1)
            self.assertEqual(3, index1.getCollectionStatistics().getNumberOfDocuments())

            index2 = pt.IndexFactory.of(indexref2)
            self.assertEqual(1, index2.getCollectionStatistics().getNumberOfDocuments())

            index_combined = pt.java.autoclass("org.terrier.python.IndexWithBackground")(index2, index1)
            self.assertEqual(3, index_combined.getCollectionStatistics().getNumberOfDocuments())

            self.assertEqual(1, index_combined.getLexicon()["test"].getFrequency())

            # this is 1 as we used the background index for the background
            # WITHOUT adding the statistics of the local index
            self.assertEqual(1, index_combined.getLexicon()["wave"].getFrequency())
            

        except JavaException as ja:
            print(ja.stacktrace)
            raise ja
        
        

        


import unittest
import pandas as pd
import pyterrier as pt
from .base import BaseTestCase

class TestIndexer(BaseTestCase):
    def test_transform_impl_none(self):
        class TestIndexer(pt.Indexer):
            pass
        indexer = TestIndexer()
        with self.assertRaises(NotImplementedError):
            indexer.transform(pd.DataFrame())
        with self.assertRaises(NotImplementedError):
            indexer.transform_iter([])
        with self.assertRaises(NotImplementedError):
            indexer([])

    def test_transform_impl_transform(self):
        class TestIndexer(pt.Indexer):
            def transform(self, inp):
                return inp
        indexer = TestIndexer()
        inp = pd.DataFrame()
        pd.testing.assert_frame_equal(inp, indexer.transform(inp))
        inp = []
        self.assertEqual([], list(indexer.transform_iter(inp)))
        self.assertEqual([], indexer(inp))

    def test_transform_impl_transform_iter(self):
        class TestIndexer(pt.Indexer):
            def transform_iter(self, inp):
                return inp
        indexer = TestIndexer()
        inp = pd.DataFrame()
        pd.testing.assert_frame_equal(inp, indexer.transform(inp))
        inp = []
        self.assertEqual(inp, indexer.transform_iter(inp))
        self.assertEqual([], indexer(inp))

    def test_transform_impl_both(self):
        class TestIndexer(pt.Indexer):
            def transform_iter(self, inp):
                return inp
            def transform(self, inp):
                return inp
        indexer = TestIndexer()
        inp = pd.DataFrame()
        pd.testing.assert_frame_equal(inp, indexer.transform(inp))
        inp = []
        self.assertEqual(inp, indexer.transform_iter(inp))
        self.assertEqual([], indexer(inp))


if __name__ == "__main__":
    unittest.main()



from functools import partial
from typing import Callable, Any, Dict, Union, Optional, Sequence, Literal, List, cast
import numpy.typing as npt
import pandas as pd
import pyterrier as pt
from pyterrier.apply_base import ApplyDocumentScoringTransformer, ApplyQueryTransformer, ApplyDocFeatureTransformer, ApplyForEachQuery, ApplyIterForEachQuery, ApplyGenericTransformer, ApplyGenericIterTransformer, ApplyIndexer, DropColumnTransformer, ApplyByRowTransformer, RenameColumnsTransformer

def _bind(instance, func, as_name=None):
    """
    Bind the function *func* to *instance*, with either provided name *as_name*
    or the existing name of *func*. The provided *func* should accept the 
    instance as the first argument, i.e. "self".
    """
    if as_name is None:
        as_name = func.__name__
    bound_method = func.__get__(instance, instance.__class__)
    setattr(instance, as_name, bound_method)
    return bound_method

def query(fn : Callable[[Union[pd.Series,pt.model.IterDictRecord]], str], *args, required_columns : Optional[List[str]] = ['qid', 'query'], **kwargs) -> pt.Transformer:
    """
        Create a transformer that takes as input a query, and applies a supplied function to compute a new query formulation.

        The supplied function is called once for each query, and must return a string containing the new query formulation.
        Each time it is called, the function is supplied with a Panda Series or dict representing the attributes of the query.
        The particular type of the input is not controlled by the implementor, so the function should be written to support 
        both, e.g. using ``row["key"]`` notation and not the ``row.key`` that is supported by a Series.

        The previous query formulation is saved in the "query_0" column. If a later pipeline stage is intended to resort to
        be executed on the previous query formulation, a ``pt.rewrite.reset()`` transformer can be applied.  

        :param fn: the function to apply to each row. It must return a string containing the new query formulation.
        :param required_columns: The list of columns that must be present in the input dataframe. Defaults to ['qid', 'query'].
        :param verbose: if set to True, a TQDM progress bar will be displayed

        Examples::

            # this will remove pre-defined stopwords from the query
            stops=set(["and", "the"])

            # a naieve function to remove stopwords - takes as input a Pandas Series, and returns a string
            def _remove_stops(q):
                terms = q["query"].split(" ")
                terms = [t for t in terms if not t in stops ]
                return " ".join(terms)

            # a query rewriting transformer that applies the _remove_stops to each row of an input dataframe
            p1 = pt.apply.query(_remove_stops) >> pt.terrier.Retriever(index, wmodel="DPH")

            # an equivalent query rewriting transformer using an anonymous lambda function
            p2 = pt.apply.query(
                    lambda q :  " ".join([t for t in q["query"].split(" ") if t not in stops ])
                ) >> pt.terrier.Retriever(index, wmodel="DPH")

        In both of the example pipelines above (`p1` and `p2`), the exact topics are not known until the pipeline is invoked, e.g.
        by using `p1.transform(topics)` on a topics dataframe, or within a ``pt.Experiment()``. When the pipeline 
        is invoked, the specified function (`_remove_stops` in the case of `p1`) is called for **each** row of the 
        input datatrame (becoming the `q` function argument).
            

    """
    return ApplyQueryTransformer(fn, *args, required_columns=required_columns, **kwargs)

def doc_score(fn : Union[Callable[[Union[pd.Series,pt.model.IterDictRecord]], float], Callable[[pd.DataFrame], Sequence[float]]], *args, required_columns : Optional[List[str]] = ['qid', 'query', 'docno'], batch_size=None, **kwargs) -> pt.Transformer:
    """
        Create a transformer that takes as input a ranked documents dataframe, and applies a supplied function to compute a new score.
        Ranks are automatically computed. doc_score() can operate row-wise, or batch-wise, depending on whether batch_size is set.

        The supplied function is called once for each document, and must return a float containing the new score for that document.
        Each time it is called, the function is supplied with a Panda Series representing the attributes of the query and document.

        :param fn: the function to apply to each row
        :param batch_size: How many documents to operate on at once (batch-wise). If None, operates row-wise
        :param required_columns: If provided, should be a list of columns that must be present in the input dataframe. Defaults to ['qid', 'query', 'docno'].
        :param verbose: if set to True, a TQDM progress bar will be displayed

        Example (Row-wise)::

            # this transformer will subtract 5 from the score of each document
            p = pt.terrier.Retriever(index, wmodel="DPH") >> 
                pt.apply.doc_score(lambda doc : doc["score"] -5) # doc["score"] works for both a dict and Series

        Can be used in batch-wise manner, which is particularly useful for appling neural models. In this case,
        the scoring function receives a dataframe, rather than a single row::

            def _doclen(df):
                # returns series of lengths
                return df.text.str.len()
            
            pipe = pt.terrier.Retriever(index) >> pt.apply.doc_score(_doclen, batch_size=128)

        Can also be used to create individual features that are combined using the ``**`` feature-union operator::

            pipeline = bm25 >> ( some_features ** pt.apply.doc_score(_doclen) )

    """
    return ApplyDocumentScoringTransformer(fn, *args, batch_size=batch_size, **kwargs)

def doc_features(fn : Callable[[Union[pd.Series,pt.model.IterDictRecord]], npt.NDArray[Any]], *args, required_columns : Optional[List[str]] = ['qid', 'query', 'docno'], **kwargs) -> pt.Transformer:
    """
        Create a transformer that takes as input a ranked documents dataframe, and applies the supplied function to each document to compute feature scores. 

        The supplied function is called once for each document, must each time return a 1D numpy array.
        Each time it is called, the function is supplied with a Panda Series, or a dictionary, representing the attributes of the query and document. The
        particular type of the input is not controlled by the implementor, so the function should be written to support both, e.g. using ``row["key"]``
        notation and not the ``row.key`` that is supported by a Series.

        :param fn: the function to apply to each row. It must return a 1D numpy array
        :param required_columns: The list of columns that must be present in the input dataframe.
        :param verbose: if set to True, a TQDM progress bar will be displayed
        
        Example::

            # this transformer will compute the character and number of word in each document retrieved
            # using the contents of the document obtained from the MetaIndex

            def _features(row):
                docid = row["docid"]
                content = index.getMetaIndex().getItem("text", docid)
                f1 = len(content)
                f2 = len(content.split(" "))
                return np.array([f1, f2])

            p = pt.terrier.Retriever(index, wmodel="BM25") >> 
                pt.apply.doc_features(_features )

        NB: If you only want to calculate a single feature to add to existing features, it is better to use ``pt.apply.doc_score()`` 
        and the ``**`` feature union operator::

            pipeline = bm25 >> ( some_features ** pt.apply.doc_score(one_feature) )

    """
    return ApplyDocFeatureTransformer(fn, *args, required_columns = required_columns, **kwargs)

def indexer(fn : Callable[[pt.model.IterDict], Any], required_columns: Optional[List[str]] = None, **kwargs) -> pt.Indexer:
    """
        Create an instance of pt.Indexer using a function that takes as input an interable dictionary.

        The supplied function is called once. It may optionally return something (typically a reference to the "index").

        :param fn: the function that consumes documents as IterDicts.
        :param required_columns: If provided, should be a list of columns that must be present in the input IterDicts.

        Example::

            # make a pt.Indexer that returns the numnber of documents consumed
            def _counter(iter_dict):
                count = 0
                for d in iter_dict:
                    count += 1
                return count
            indexer = pt.apply.indexer(_counter)
            rtr = indexer.index([ {'docno' : 'd1'}, {'docno' : 'd2'}])
    """
    return ApplyIndexer(fn, required_columns = required_columns, **kwargs)

def rename(columns: Dict[str,str], *, errors: Literal['raise', 'ignore'] = 'raise') -> pt.Transformer:
    """
        Creates a transformer that renames columns in a dataframe. 

        :param columns: A dictionary mapping from old column name to new column name
        :param errors: Maps to df.rename() errors kwarg - default to 'raise', alternatively can be 'ignore'

        Example::
            
            pipe = pt.terrier.Retriever(index, metadata=["docno", "body"]) >> pt.apply.rename({'body':'text'})
    """
    return RenameColumnsTransformer(columns, errors=errors)

def generic(
        fn : Union[Callable[[pd.DataFrame], pd.DataFrame], Callable[[pt.model.IterDict], pt.model.IterDict]], 
        *args, 
        batch_size : Optional[int] = None, 
        iter : bool = False, 
        transform_outputs : Optional[Callable[[List[str]], List[str]]] = None, 
        **kwargs) -> pt.Transformer:
    """
        Create a transformer that changes the input dataframe to another dataframe in an unspecified way.

        The supplied function is called once for an entire result set as a dataframe or iter-dict (which may contain one or 
        more queries and one or more documents). Each time it should return a new dataframe. The returned dataframe (or yielded row) 
        should abide by the general PyTerrier Data Model, for instance updating the rank column if the scores are amended.

        :param fn: the function to apply to each result set
        :param batch_size: whether to apply fn on batches of rows or all that are received
        :param required_columns: If provided, should be a list of columns that must be present in the input dataframe.
        :param verbose: Whether to display a progress bar over batches (only used if batch_size is set, and iter is not set).
        :param iter: Whether to use the iter-dict API - if-so, then ``fn`` receives an iterable, and returns an iterable. 
        :param transform_outputs: Used to support inspection. If provided, should be a function (e.g. lambda) that takes 
            the input columns as an argument, and returns the list of columns that the transformer will output. 
            This need only be set if you need inspectability and iter=True, or your transformer doesn't respond well to 
            being inspected by empty dataframes.

        Example (dataframe)::

            # this pipeline would remove all but the first two documents from a result set
            pipe = pt.terrier.Retriever(index) >> pt.apply.generic(lambda res : res[res["rank"] < 2])

        Example (iter-dict)::

            # this pipeline would simlarly remove all but the first two documents from a result set
            def _fn(iterdict):
                for result in iterdict:
                    if result["rank"] < 2:
                        yield result
            pipe1 = pt.terrier.Retriever(index) >> pt.apply.generic(_fn, iter=True)

            # transform_iter() can also return an iterable, so returning a list is also permissible
            pipe2 = pt.terrier.Retriever(index) >> pt.apply.generic(lambda res: [row for row in res if row["rank"] < 2], iter=True)

    """
    rtr : pt.Transformer
    if iter:
        if kwargs.get("add_ranks", False):
            raise ValueError("add_ranks=True not supported with iter=True")
        fn = cast(Callable[[pt.model.IterDict], pt.model.IterDict], fn) # noqa: PT100 (this is typing.cast, not jinus.cast)
        rtr = ApplyGenericIterTransformer(fn, *args, batch_size=batch_size, **kwargs)
    else:
        fn = cast(Callable[[pd.DataFrame], pd.DataFrame], fn) # noqa: PT100 (this is typing.cast, not jinus.cast)
        rtr = ApplyGenericTransformer(fn, *args, batch_size=batch_size, **kwargs)
    if transform_outputs is not None:
        rtr.transform_outputs = transform_outputs # type: ignore[attr-defined]
    return rtr

def by_query(
        fn : Union[Callable[[pd.DataFrame], pd.DataFrame], Callable[[pt.model.IterDict], pt.model.IterDict]], 
        *args, 
        batch_size : Optional[int] = None, 
        iter : bool = False, 
        verbose : bool = False,
        transform_outputs : Optional[Callable[[List[str]], List[str]]] = None,  
        **kwargs) -> pt.Transformer:
    """
        As `pt.apply.generic()` except that fn receives a dataframe (or iter-dict) for one query at at time, rather than all results at once.
        If batch_size is set, fn will receive no more than batch_size documents for any query. The verbose kwargs controls whether
        to display a progress bar over queries.  

        :param fn: the function to apply to each row. Should return a generator
        :param batch_size: whether to apply fn on batches of rows or all that are received.
        :param required_columns: If provided, should be a list of columns that must be present in the input dataframe.
        :param verbose: Whether to display a progress bar over batches (only used if batch_size is set, and iter is not set).
        :param iter: Whether to use the iter-dict API - if-so, then ``fn`` receives an iterable, and must return an iterable. 
        :param transform_outputs: Used to support inspection. If provided, should be a function (e.g. lambda) that takes 
            the input columns as an argument, and returns the list of columns that the transformer will output. 
            This need only be set if you need inspectability and iter=True, or your transformer doesn't respond well to 
            being inspected by empty dataframes.
    """
    rtr : pt.Transformer
    if iter:
        fn = cast(Callable[[pt.model.IterDict], pt.model.IterDict], fn) # noqa: PT100 (this is typing.cast, not jinus.cast)
        if kwargs.get("add_ranks", False):
            raise ValueError("add_ranks=True not supported with iter=True")
        rtr = ApplyIterForEachQuery(fn, *args, batch_size=batch_size, verbose=verbose, **kwargs)
    else:
        fn = cast(Callable[[pd.DataFrame], pd.DataFrame], fn) # noqa: PT100 (this is typing.cast, not jinus.cast)
        rtr = ApplyForEachQuery(fn, *args, batch_size=batch_size, verbose=verbose, **kwargs)
    if transform_outputs is not None:
        rtr.transform_outputs = transform_outputs # type: ignore[attr-defined]
    return rtr

class _apply:

    def __init__(self):
        _bind(self, lambda self, fn, *args, **kwargs : query(fn, *args, **kwargs), as_name='query')
        _bind(self, lambda self, fn, *args, **kwargs : doc_score(fn, *args, **kwargs), as_name='doc_score')
        _bind(self, lambda self, fn, *args, **kwargs : doc_features(fn, *args, **kwargs), as_name='doc_features')
        _bind(self, lambda self, fn, *args, **kwargs : indexer(fn, *args, **kwargs), as_name='indexer')
        _bind(self, lambda self, fn, *args, **kwargs : rename(fn, *args, **kwargs), as_name='rename')
        _bind(self, lambda self, fn, *args, **kwargs : by_query(fn, *args, **kwargs), as_name='by_query')
        _bind(self, lambda self, fn, *args, **kwargs : generic(fn, *args, **kwargs), as_name='generic')     
    
    def __getattr__(self, item: str) -> Callable[..., pt.Transformer]:
        return partial(generic_apply, item)

def generic_apply(
    name: str,
    fn=None,
    *,
    drop: bool = False,
    batch_size: Optional[int] = None,
    required_columns: Optional[List[str]] = None,
    verbose=False
) -> pt.Transformer:
    if drop:
        assert fn is None, "cannot provide both fn and drop=True"
        return DropColumnTransformer(name)

    return ApplyByRowTransformer(name, fn, batch_size=batch_size, required_columns=required_columns, verbose=verbose)



from copy import copy
import html
import uuid
from importlib import resources
from typing import Any, Dict, List, Optional, Protocol, Union, cast, runtime_checkable

import numpy as np
import pyterrier as pt
import re

def _apply_default_schematic(schematic: Dict[str, Any], transformer: pt.Transformer, *, input_columns: Optional[List[str]] = None):
    schematic.setdefault('type', 'indexer' if pt.inspect.transformer_type(transformer) == pt.inspect.TransformerType.indexer else 'transformer')
    assert schematic['type'] in ('transformer', 'indexer')

    if 'label' not in schematic:
        label = transformer.__class__.__name__
        if label.endswith('Transformer'):
            schematic['label'] = label[:-len('Transformer')]
        else:
            schematic['label'] = label

    if 'class_name' not in schematic:
        name = f'{transformer.__class__.__module__}.{transformer.__class__.__name__}'
        if name.startswith('pyterrier.'):
            name = 'pt.' + name[len('pyterrier.'):]
        schematic['name'] = name

    if 'input_columns' not in schematic or 'output_columns' not in schematic:
        if 'input_columns' not in schematic:
            schematic['input_columns'] = input_columns
        if 'output_columns' not in schematic:
            if input_columns is None:
                schematic['output_columns'] = None
            else:
                try:
                    schematic['output_columns'] = pt.inspect.transformer_outputs(transformer, input_columns)
                except pt.validate.InputValidationError as e:
                    schematic['output_columns'] = None
                    schematic['input_validation_error'] = e
                except pt.inspect.InspectError:
                    schematic['output_columns'] = None

    default_settings_applied = False
    if 'settings' not in schematic:
        default_settings_applied = True
        try:
            schematic['settings'] = {attr.name: attr.value for attr in pt.inspect.transformer_attributes(transformer)}
        except pt.inspect.InspectError:
            schematic['settings'] = {}

    if 'help_url' not in schematic:
        schematic['help_url'] = pt.documentation.url_for_class(transformer)

    if 'inner_pipelines' not in schematic:
        try:
            subtransformers = pt.inspect.subtransformers(transformer)
        except pt.inspect.InspectError:
            subtransformers = {}
        if subtransformers:
            subtransformer_inputs = schematic['input_columns'] or _INFER
            if schematic.get('inner_pipelines_mode', 'unlinked') == 'unlinked':
                subtransformer_inputs = _INFER
            pipelines = []
            pipeline_labels = []
            for key, value in subtransformers.items():
                if default_settings_applied and key in schematic['settings']:
                    del schematic['settings'][key]
                if isinstance(value, list):
                    for i, v in enumerate(value):
                        pipelines.append(transformer_schematic(v, input_columns=subtransformer_inputs))
                        pipeline_labels.append(f'{key}[{i}]')
                else:
                    pipelines.append(transformer_schematic(value, input_columns=subtransformer_inputs))
                    pipeline_labels.append(key)
            schematic['inner_pipelines'] = pipelines
            schematic['inner_pipelines_labels'] = pipeline_labels

    if schematic.get('inner_pipelines') and 'inner_pipelines_mode' not in schematic:
        schematic['inner_pipelines_mode'] = 'unlinked'


_INFER = object()
def transformer_schematic(
    transformer: pt.Transformer,
    *,
    input_columns: Optional[Union[List[str],object]] = _INFER,
    default: bool = False,
) -> dict:
    """Builds a structured schematic of the transformer."""
    is_indexer = pt.inspect.transformer_type(transformer) == pt.inspect.TransformerType.indexer
    if input_columns is _INFER:
        if is_indexer:
            all_input_column_configs = pt.inspect.indexer_inputs(cast(pt.Indexer, transformer), strict=False) # noqa: PT100
        else:
            all_input_column_configs = pt.inspect.transformer_inputs(transformer, strict=False)
        if all_input_column_configs is not None and len(all_input_column_configs) > 0:
            input_columns = all_input_column_configs[0] # pick the first one
        else:
            input_columns = None
    # input_columns can no longer be _INFER
    input_columns = cast(Optional[List[str]], input_columns) # noqa: PT100 (this is typing.cast, not jinus.cast)
    if not default and isinstance(transformer, HasSchematic):
        if callable(transformer.schematic):
            schematic = transformer.schematic(input_columns=input_columns)
        else:
            schematic = transformer.schematic
        schematic = copy(schematic) # we don't want to accidently modify the original schematic
        if 'type' in schematic:
            return schematic
    else:
        schematic = {}
    _apply_default_schematic(schematic, transformer, input_columns=input_columns)
    return schematic

# Tools for converting the schematic diagrams to html
_css = None
_js = None
def _get_schematic_css_js(container_id):
    global _css, _js
    if _css is None or _js is None:
        _css = (resources.files('pyterrier') / 'data/schematic.css').read_text()
    if _js is None:
        _js = (resources.files('pyterrier') / 'data/schematic.js').read_text()
    css = _css.replace('#ID', f'#{container_id}')
    js = _js.replace('#ID', f'#{container_id}')
    return css, js


def draw(transformer: Union[pt.Transformer, dict], *, 
         outer_class: Optional[str] = None, 
         input_columns: Optional[List[str]] = None) -> str:
    """Draws a transformer as an HTML schematic.

    If the transformer is already a ``SchematicDict``, it will be drawn directly.
    Otherwise, it will first convert the transformer to a structured schematic using :func:`transformer_schematic`,
    and draw that.

    Args:
        transformer: The transformer to draw, or a dict in ``SchematicDict`` format.
        input_columns: If you want to specify the input columns for the transformer (pipeline).
        outer_class: An optional CSS class to apply to the outer container of the schematic.

    Returns:
        An HTML string representing the schematic of the transformer.
    """
    if isinstance(transformer, dict):
        assert input_columns is None, "Cannot set input_columns and provide a SchematicDict input."
        schematic = transformer
    else:
        schematic = transformer_schematic(transformer, input_columns=input_columns or _INFER)
    return draw_html_schematic(schematic, outer_class=outer_class)


def draw_html_schematic(schematic: dict, *, outer_class: Optional[str] = None) -> str:
    """Draws a structured schematic as an HTML representation."""
    uid = str(uuid.uuid4())
    css, js = _get_schematic_css_js(f'id-{uid}')
    return f'''
    <div id="id-{uid}" class="{outer_class or ''}" style="display: none;">
        <style>{css}</style>
        {_draw_html_schematic(schematic)}
        <div class="pts-infobox">
            <div class="pts-infobox-title"></div>
            <div class="pts-infobox-body"></div>
            <div class="pts-infobox-hint"><svg  xmlns="http://www.w3.org/2000/svg"  width="24"  height="24"  viewBox="0 0 24 24"  fill="currentColor"  style="width: 12px; height: 12px; vertical-align: -2px;"><path stroke="none" d="M0 0h24v24H0z" fill="none"/><path d="M4 11a1 1 0 0 1 .117 1.993l-.117 .007h-1a1 1 0 0 1 -.117 -1.993l.117 -.007h1z" /><path d="M12 2a1 1 0 0 1 .993 .883l.007 .117v1a1 1 0 0 1 -1.993 .117l-.007 -.117v-1a1 1 0 0 1 1 -1z" /><path d="M21 11a1 1 0 0 1 .117 1.993l-.117 .007h-1a1 1 0 0 1 -.117 -1.993l.117 -.007h1z" /><path d="M4.893 4.893a1 1 0 0 1 1.32 -.083l.094 .083l.7 .7a1 1 0 0 1 -1.32 1.497l-.094 -.083l-.7 -.7a1 1 0 0 1 0 -1.414z" /><path d="M17.693 4.893a1 1 0 0 1 1.497 1.32l-.083 .094l-.7 .7a1 1 0 0 1 -1.497 -1.32l.083 -.094l.7 -.7z" /><path d="M14 18a1 1 0 0 1 1 1a3 3 0 0 1 -6 0a1 1 0 0 1 .883 -.993l.117 -.007h4z" /><path d="M12 6a6 6 0 0 1 3.6 10.8a1 1 0 0 1 -.471 .192l-.129 .008h-6a1 1 0 0 1 -.6 -.2a6 6 0 0 1 3.6 -10.8z" /></svg> Click to explore!</div>
        </div>
        <script>{js}</script>
    </div>
    <div id="id-{uid}-pts-rendering-issue">
        Rendering isssue. Try running the cell again.
    </div>
    '''


def _draw_html_schematic(schematic: dict, *, mode: str = 'outer') -> str:
    if schematic['type'] == 'transformer':
        return _draw_html_schematic({
            'type': 'pipeline',
            'input_columns': schematic.get('input_columns'),
            'output_columns': schematic.get('output_columns'),
            'transformers': [schematic],
        }, mode=mode)
    if schematic['type'] == 'indexer':
        return _draw_html_schematic({
            'type': 'pipeline',
            'input_columns': schematic.get('input_columns'),
            'transformers': [schematic],
        }, mode=mode)
    if schematic['type'] == 'pipeline':
        result = '<div class="pts-pipeline">'
        if mode == 'outer':
            clz = 'pts-arr' if schematic['transformers'][0].get('inner_pipelines_mode') != 'combine' else ''
            result += '<div class="pts-io-label">Input</div>'
            result += f'<div class="pts-hline {clz} pts-arr-input">{_draw_df_html(schematic["input_columns"])}</div>'
        elif mode == 'inner_linked':
            result += '<div class="pts-hline pts-arr pts-arr-inner" style="width: 16px;"></div>'
        elif mode == 'inner_labeled':
            result += f'<div class="pts-hline pts-arr pts-arr-input">{_draw_df_html(schematic["input_columns"])}</div>'
        columns = schematic["input_columns"]
        for i, record in enumerate(schematic['transformers']):
            assert record['type'] == 'transformer' or record['type'] == 'indexer'
            assert record['input_columns'] == columns
            uid = str(uuid.uuid4())
            infobox = ''
            infobox_attr = ''
            error_cls = ''
            if record.get('settings') or record.get('name') or record.get('input_validation_error'):
                help_url = record.get('help_url') or ''
                name = record.get('name') or ''
                attrs = ''
                error_info = ''
                if record.get('input_validation_error'):
                    modes = record['input_validation_error'].modes
                    error_cls = 'pts-input-validation-error'
                    if len(modes) == 1:
                        # Normal case: there's just one mode
                        error_info = '<div class="pts-infobox-error">'
                        if len(modes[0].missing_columns) > 0:
                            error_info += f'Missing input columns: {", ".join(["<b>" + html.escape(c) + "</b>" for c in modes[0].missing_columns])}. '
                        if len(modes[0].extra_columns) > 0:
                            error_info += f'Unexpected input columns: {", ".join(["<b>" + html.escape(c) + "</b>" for c in modes[0].extra_columns])}. '
                        error_info += '</div>'
                    else:
                        error_info = '<div class="pts-infobox-error"><div>None of the supported input modes matched:</div><ul>'
                        for i, error_mode in enumerate(modes):
                            error_info += f'<li>Mode {html.escape(error_mode.mode_name or str(i+1))}: '
                            if len(error_mode.missing_columns) > 0:
                                error_info += f'Missing input columns: {", ".join(["<b>" + html.escape(c) + "</b>" for c in error_mode.missing_columns])}. '
                            if len(error_mode.extra_columns) > 0:
                                error_info += f'Unexpected input columns: {", ".join(["<b>" + html.escape(c) + "</b>" for c in error_mode.extra_columns])}. '
                            error_info += '</li>'
                        error_info += '</ul></div>'
                if record['settings']:
                    attr_rows = []
                    for key, value in record['settings'].items():
                        attr_rows.append(f'<tr><th>{html.escape(key)}</th><td>{html.escape(str(value))}</td></tr>')
                    attrs = f'<table class="pts-df-columns">{"".join(attr_rows)}</table>'
                infobox = f'''
                <div class="pts-infobox-item" id="id-{uid}" data-title="Transformer">
                    <div style="font-family: monospace; padding: 3px 6px;">
                        {'<a href="' + html.escape(help_url) + '" target="_blank" onclick="window.event.stopPropagation();">' if help_url else ''}
                            {html.escape(name)}
                        {'</a>' if help_url else ''}
                    </div>
                    {error_info}
                    {attrs}
                </div>
                '''
                infobox_attr = f'data-pts-infobox="id-{uid}"'
            if 'inner_pipelines' in record:
                if record['inner_pipelines_mode'] == 'linked':
                    pipelines = ''
                    for pipeline in record['inner_pipelines']:
                        pipelines += '<div class="pts-parallel-item"><div class="pts-vline"></div>' + _draw_html_schematic(pipeline, mode='inner_linked') + '<div class="pts-vline"></div></div>'
                    result += f'''
                    <div class="pts-transformer pts-inner pts-parallel-scaffold {error_cls}" {infobox_attr}>
                        {infobox}
                        <div class="pts-hline"></div>
                        <div class="pts-transformer-title">{html.escape(record["label"])}</div>
                        <div class="pts-inner-schematic pts-inner-linked">{pipelines}</div>
                        <div class="pts-hline pts-arr"></div> <!-- TODO this is unusual - an arrow AFTER something -->
                    </div>
                    '''
                elif record['inner_pipelines_mode'] == 'combine':
                    pipelines = ''
                    for pipeline in record['inner_pipelines']:
                        pipelines += '<div class="pts-parallel-item"><div class="pts-vline"></div>' + _draw_html_schematic(pipeline, mode='inner_linked') + '<div class="pts-vline"></div></div>'
                    result += f'''
                    <div class="pts-combine-box">
                        <div class="pts-parallel-scaffold pts-inner">
                            <div class="pts-hline"></div>
                            <div class="pts-inner-schematic pts-inner-linked">{pipelines}</div>
                            <div class="pts-hline pts-arr"></div>
                        </div>
                        <div class="pts-transformer {error_cls}" {infobox_attr}>
                            {infobox}
                            <div class="pts-transformer-title">{html.escape(record["label"])}</div>
                        </div>
                    </div>
                    '''
                elif record['inner_pipelines_mode'] == 'unlinked':
                    pipelines = ''
                    if len(record['inner_pipelines_labels']) == len(record['inner_pipelines']):
                        for label, pipeline in zip(record['inner_pipelines_labels'], record['inner_pipelines']):
                            pipelines += f'''
                            <div class="pts-transformer-title">{html.escape(label)}</div>
                            <div class="pts-inner-schematic pts-inner-labeled">{_draw_html_schematic(pipeline, mode='inner_labeled')}</div>
                            '''
                    else:
                        for pipeline in record['inner_pipelines']:
                            pipelines += _draw_html_schematic(pipeline, mode='inner_labeled')
                    result += f'''
                    <div class="pts-transformer pts-inner {error_cls}" {infobox_attr}>
                        {infobox}
                        <div class="pts-transformer-title">{html.escape(record["label"])}</div>
                        <div class="pts-inner-schematic pts-inner-labeled">{pipelines}</div>
                    </div>
                    '''
            elif record['type'] == 'indexer':
                result += f'''
                <div class="pts-transformer {error_cls}" {infobox_attr}>
                    {infobox}
                    <div class="pts-transformer-title">{html.escape(record["label"])}</div>
                </div>
                '''
            else:
                assert record['type'] == 'transformer'
                result += f'''
                <div class="pts-transformer {error_cls}" {infobox_attr}>
                    {infobox}
                    <div class="pts-transformer-title">{html.escape(record["label"])}</div>
                </div>
                '''
            if i != len(schematic['transformers']) - 1:
                result += f'<div class="pts-hline pts-arr pts-arr-inner">{_draw_df_html(record["output_columns"], record["input_columns"])}</div>'
            columns = record['output_columns']
        if mode == 'outer':
            if schematic["transformers"][-1]["type"] == 'indexer':
                result += '<div class="pts-hline pts-arr pts-arr-output"><svg xmlns="http://www.w3.org/2000/svg"  width="24"  height="24"  viewBox="0 0 24 24"  fill="none"  stroke="currentColor"  stroke-width="2"  stroke-linecap="round"  stroke-linejoin="round" class="pts-artifact-icon"><path stroke="none" d="M0 0h24v24H0z" fill="none"/><path d="M12 6m-8 0a8 3 0 1 0 16 0a8 3 0 1 0 -16 0" /><path d="M4 6v6a8 3 0 0 0 16 0v-6" /><path d="M4 12v6a8 3 0 0 0 16 0v-6" /></svg></div>'
                result += '<div class="pts-io-label">Output</div>'
            else:
                result += f'<div class="pts-hline pts-arr pts-arr-output">{_draw_df_html(schematic["output_columns"], schematic["transformers"][-1]["input_columns"])}</div>'
                result += '<div class="pts-io-label">Output</div>'
        elif mode == 'inner_linked':
            result += f'<div class="pts-hline" style="flex-grow: 1;">{_draw_df_html(schematic["output_columns"], schematic["transformers"][-1]["input_columns"])}</div>'
        elif mode == 'inner_labeled':
            result += f'<div class="pts-hline pts-arr pts-arr-output">{_draw_df_html(schematic["output_columns"], schematic["transformers"][-1]["input_columns"])}</div>'
        result += '</div>'
        return result
    raise ValueError(f"Unknown schematic type {schematic['type']}")


def _draw_df_html(columns, prev_columns = None) -> str:
    """Draws a DataFrame as an HTML table."""
    if columns is None:
        columns = []
        df_class = ' pts-df-alert'
        frame_info = {
            'label': '?',
            'title': 'Unknown Frame',
        }
    else:
        df_class = ''
        frame_info = pt.model.frame_info(columns) or {'label': 'DF', 'title': 'DataFrame'}
    df_label = frame_info['label'] 
    df_label_long = frame_info['title'] 
    # change underscore subscript into HTML subscript
    df_label = re.sub(r'_(\w+)', r'<sub>\1</sub>', df_label)
    uid = str(uuid.uuid4())
    if columns:
        column_rows = []
        for col in columns:
            col_info = pt.model.column_info(col) or {}
            col_desc = ''
            type_name = ''
            if 'type' in col_info:
                type_name = str(col_info['type'])
                if col_info['type'] == np.array:
                    type_name = 'np.array'
                elif hasattr(col_info['type'], '__name__'):
                    type_name = col_info['type'].__name__
                type_name = f'<span style="font-family: monospace;">{html.escape(type_name)}</span>'
            if 'phrase' in col_info:
                col_desc += f'<i>({html.escape(col_info["phrase"])})</i> '
            if 'short_desc' in col_info:
                col_desc += f'{html.escape(col_info["short_desc"])} '
            is_added = prev_columns and col not in prev_columns
            column_rows.append(f'''
                <tr class="{"pts-add" if is_added else ""}">
                    <th>{html.escape(col)}</th>
                    <td>{type_name}</td>
                    <td>{col_desc}</td>
                </tr>
            ''')
        col_table = f'''
        <div id="id-{uid}" class="pts-infobox-item" data-title="{df_label_long}">
            <table class="pts-df-columns">
                {''.join(column_rows)}
            </table>
        </div>'''
    else:
        col_table = f'''
        <div id="id-{uid}" class="pts-infobox-item" data-title="{df_label_long}">
            <div class="pts-infobox-error">Unknown/incompatible columns</div>
        </div>'''
    return f'<div class="pts-df {df_class}" data-pts-infobox="id-{uid}">{df_label}{col_table}</div>'


@runtime_checkable
class HasSchematic(Protocol):
    """Protocol for transformers override details about their schematic representation.

    This is an optional extension interface to :class:`pyterrier.Transformer` that allows
    transformers to provide customizations to their schematics.
    """
    def schematic(self, *, input_columns: Optional[List[str]]) -> Dict[str, Any]:
        """Returns a structured schematic representation of the transformer.

        The schematic should be a dictionary that follows the structure defined in :ref:`pt.schematic <pyterrier.schematic>`.

        For ease of use, the method can optionally return only some of the fields of the schematic; any missing fields
        will be filled in with default values.

        It can also be implemented as an instance or class member when the values do not need to be computed on-the-fly (e.g.,
        overriding the schematic label). When ``schematic`` is not ``callable``, it uses its dict value directly as the schematic.

        Args:
            input_columns: The input columns of the transformer, used to determine schematic fields such as the output columns.

        Returns:
            A dictionary representing the schematic of the transformer, which will be used to draw the schematic diagram.
        """



import pyterrier as pt

import unittest
import os
from .base import TempDirTestCase, ensure_deprecated


class TestDFIndexer(TempDirTestCase):

    @ensure_deprecated
    def _create_index(self, type, dfText, dfMeta):
        print("Writing index type "+str(type)+" to " + self.test_dir)
        pd_indexer = pt.DFIndexer(self.test_dir, type=type)
        indexref = pd_indexer.index(dfText, dfMeta)
        self.assertIsNotNone(indexref)
        return indexref

    def _make_check_index(self, n, index_type, include_urls=False):
        from pyterrier.terrier.index import IndexingType
        import pandas as pd
        df = pd.DataFrame({
            'docno': ['1', '2', '3'],
            'url':
                ['url1', 'url2', 'url3'],
            'text':
                ['He ran out of money, so he had to stop playing',
                 'The waves were crashing on the shore; it was a',
                 'The body may perhaps compensates for the loss']
        })
        df = df.head(n)
        metadata = df[["docno", "url"]] if include_urls else df["docno"]
        indexref = self._create_index(index_type, df["text"], metadata)
        index = pt.IndexFactory.of(indexref)
        self.assertIsNotNone(index)
        self.assertEqual(n, index.getCollectionStatistics().getNumberOfDocuments())
        self.assertTrue("docno" in index.getMetaIndex().getKeys())
        if include_urls:
            self.assertTrue("url" in index.getMetaIndex().getKeys())
        if index_type is IndexingType.CLASSIC:
            self.assertTrue(os.path.isfile(self.test_dir + '/data.direct.bf'))
            self.assertTrue(index.hasIndexStructure("direct"))
        else:
            self.assertFalse(os.path.isfile(self.test_dir + '/data.direct.bf'))

    def test_checkjavaDocIteratr_None(self):
        import pandas as pd
        df = pd.DataFrame({
            'docno':
                ['1', '2', '3'],
            'url':
                ['url1', 'url2', 'url3'],
            'text':
                [None,
                 'The waves were crashing on the shore; it was a',
                 'Some other text']
        })

        from pyterrier import DFIndexUtils

        d1 = df.head(1)
        jIter1, metalens = DFIndexUtils.create_javaDocIterator(d1["text"], d1["docno"])
        self.assertEqual(1, len(metalens))
        self.assertEqual(1, metalens["docno"])
        self.assertTrue(jIter1.hasNext())
        self.assertIsNotNone(jIter1.next())
        self.assertFalse(jIter1.hasNext())

        d2 = df.head(2)
        jIter1, metalens = DFIndexUtils.create_javaDocIterator(d2["text"], d2[["docno", "url"]])
        self.assertEqual(2, len(metalens))
        self.assertEqual(1, metalens["docno"])
        self.assertEqual(4, metalens["url"])
        self.assertTrue(jIter1.hasNext())
        self.assertIsNotNone(jIter1.next())
        self.assertTrue(jIter1.hasNext())
        self.assertIsNotNone(jIter1.next())
        self.assertFalse(jIter1.hasNext())

    def test_checkjavaDocIteratr(self):
        import pandas as pd
        df = pd.DataFrame({
            'docno':
                ['1', '2', '3'],
            'url':
                ['url1', 'url2', 'url3'],
            'text':
                ['He ran out of money, so he had to stop playing',
                 'The waves were crashing on the shore; it was a',
                 'The body may perhaps compensates for the loss']
        })

        from pyterrier import DFIndexUtils

        d1 = df.head(1)
        jIter1, metalens = DFIndexUtils.create_javaDocIterator(d1["text"], d1["docno"])
        self.assertEqual(1, len(metalens))
        self.assertEqual(1, metalens["docno"])
        self.assertTrue(jIter1.hasNext())
        self.assertIsNotNone(jIter1.next())
        self.assertFalse(jIter1.hasNext())

        d2 = df.head(2)
        jIter1, metalens = DFIndexUtils.create_javaDocIterator(d2["text"], d2[["docno", "url"]])
        self.assertEqual(2, len(metalens))
        self.assertEqual(1, metalens["docno"])
        self.assertEqual(4, metalens["url"])
        self.assertTrue(jIter1.hasNext())
        self.assertIsNotNone(jIter1.next())
        self.assertTrue(jIter1.hasNext())
        self.assertIsNotNone(jIter1.next())
        self.assertFalse(jIter1.hasNext())

    @ensure_deprecated
    def test_badinvocation(self):
        import pandas as pd
        df_docids = pd.DataFrame([['d1', 'this is a doc']], columns=['body', 'doc_id'])
        df_docnos = df_docids.rename(columns={'doc_id' : 'docno'})
        with self.assertRaises(ValueError):
            # this should fail - there is no docno column
            ref = pt.DFIndexer(self.test_dir).index(df_docids["body"], df_docids['doc_id'])
        with self.assertRaises(ValueError):
            # this should fail - there is no docno column
            ref = pt.DFIndexer(self.test_dir).index(df_docids["body"], docn=df_docids['doc_id'])
            
        # this should pass - it picks up the metadata name from the series name
        ref = pt.DFIndexer(self.test_dir).index(df_docids["body"], df_docnos['docno'])

    def test_stopwords(self):
        import pandas as pd
        df = pd.DataFrame([{'docno': 'd1', 'text' : 'greatest hits'}])
        indexer = pt.DFIndexer(self.test_dir, stopwords=['hits'])
        indexref = indexer.index(df["text"], df["docno"])
        index = pt.IndexFactory.of(indexref)
        self.assertIsNotNone(index)
        self.assertEqual(1, index.getCollectionStatistics().getNumberOfDocuments())
        self.assertTrue("hits" not in index.getLexicon())
    
    def test_createindex1_two_metadata(self):
        from pyterrier.terrier.index import IndexingType
        self._make_check_index(1, IndexingType.CLASSIC, include_urls=True)

    def test_createindex1(self):
        from pyterrier.terrier.index import IndexingType
        self._make_check_index(1, IndexingType.CLASSIC)

    def test_createindex2(self):
        from pyterrier.terrier.index import IndexingType
        self._make_check_index(2, IndexingType.CLASSIC)

    def test_createindex3(self):
        from pyterrier.terrier.index import IndexingType
        self._make_check_index(3, IndexingType.CLASSIC)

    def test_createindex1_single_pass(self):
        from pyterrier.terrier.index import IndexingType
        self._make_check_index(1, IndexingType.SINGLEPASS)

    def test_createindex2_single_pass(self):
        from pyterrier.terrier.index import IndexingType
        self._make_check_index(2, IndexingType.SINGLEPASS)

    def test_createindex3_single_pass(self):
        from pyterrier.terrier.index import IndexingType
        self._make_check_index(3, IndexingType.SINGLEPASS)

    def test_createindex1_memory(self):
        from pyterrier.terrier.index import IndexingType
        self._make_check_index(1, IndexingType.MEMORY)

    def test_createindex2_memory(self):
        from pyterrier.terrier.index import IndexingType
        self._make_check_index(2, IndexingType.MEMORY)

    def test_createindex3_memory(self):
        from pyterrier.terrier.index import IndexingType
        self._make_check_index(3, IndexingType.MEMORY)

if __name__ == "__main__":
    unittest.main()
